## @param cood Coord from we are going to add totals
## @param list_of_totals List with strings or keys. Example: ["Total", "#SUM", "#AVG"]...
## @param styles List with string styles or None. If none tries to guest from top column object. List example: ["GrayLightPercentage", "GrayLightInteger"]
## @param string with the row where th3e total begins
## @param string with the rew where the formula ends. If None it's a coord.row -1
from unogenerator.commons import ColorsNamed, Coord as C, Range as R, guess_object_style, generate_formula_total_string
from pydicts import lod
from gettext import translation
from logging import debug
from math import ceil
from importlib.resources import files

try:
    t=translation('unogenerator', files("unogenerator") / 'locale')
    _=t.gettext
except:
    _=str

def helper_totals_row(doc, coord, list_of_totals, color=ColorsNamed.GrayLight, styles=None, row_from="2", row_to=None):
    coord=C.assertCoord(coord)
    for letter, total in enumerate(list_of_totals):
        coord_total=coord.addColumnCopy(letter)
        coord_total_from=C(coord_total.letter+row_from)
        if row_to is None:
            coord_total_to=coord_total.addRowCopy(-1)# row above
        else:
            coord_total_to=C(coord_total.letter+row_to)

        if styles is None:
            style=guess_object_style(doc.getValue(coord_total_from))
        elif styles.__class__.__name__ != "list":
            style=styles
        else:
            style=styles[letter]

        doc.addCellWithStyle(coord_total, generate_formula_total_string(total, coord_total_from, coord_total_to), color, style)


def helper_totals_column(doc, coord, list_of_totals, color=ColorsNamed.GrayLight, styles=None, column_from="B", column_to=None):
    """
        Genera una columna de totales desde la coordenada pasado como parámetro
        @param doc Documento ODS
        @param coord Coordenada de inicio
        @param list_of_totals List of values #SUM #AVG #MEDIAN
        @type list
        @param color DESCRIPTION (defaults to ColorsNamed.GrayLight)
        @type TYPE (optional)
        @param styles DESCRIPTION (defaults to None)
        @type TYPE (optional)
        @param column_from DESCRIPTION (defaults to "B")
        @type TYPE (optional)
        @param column_to DESCRIPTION (defaults to None)
        @type TYPE (optional)
    """
    coord=C.assertCoord(coord)
    for number, total in enumerate(list_of_totals):
        coord_total=coord.addRowCopy(number)
        coord_total_from=C(column_from + coord_total.number)
        if column_to is None:
            coord_total_to=coord_total.addColumnCopy(-1)# row above
        else:
            coord_total_to=C(column_to + coord_total.number)

        if styles is None:
            style=guess_object_style(doc.getValue(coord_total_from))
        elif styles.__class__.__name__ != "list":
            style=styles
        else:
            style=styles[number]

        doc.addCellWithStyle(coord_total, generate_formula_total_string(total, coord_total_from, coord_total_to), color, style)
        
def helper_title_values_total_row( doc, coord, title, values, 
        style_title=None, color_title=ColorsNamed.Orange, 
        style_values=None, color_values=ColorsNamed.White, 
        style_total=None, color_total=ColorsNamed.GrayLight
    ):
    coord=C.assertCoord(coord)

    if style_title is None:
        style_title="Bold"

    if style_total is None and len(values)>0:
        style_total=guess_object_style(values[0])


    i=0
    if title is not None:
        doc.addCellWithStyle(coord,title,color_title,style_title)
        i=i+1


    doc.addRowWithStyle(coord.addColumnCopy(i),values,colors=color_values,styles=style_values)
    doc.addCellWithStyle(coord.addColumnCopy(i+len(values)),f"=sum({coord.addColumnCopy(i).string()}:{coord.addColumnCopy(i+len(values)-1).string()}",color_total,style_total)

def helper_title_values_total_column(doc, coord, title, values,
        style_title=None, color_title=ColorsNamed.Orange, 
        style_values=None, color_values=ColorsNamed.White, 
        style_total=None, color_total=ColorsNamed.GrayLight
    ):
    coord=C.assertCoord(coord)

    if style_title is None:
        style_title="BoldCenter"

    if style_total is None and len(values)>0:
        style_total=guess_object_style(values[0])


    i=0
    if title is not None:
        doc.addCellWithStyle(coord,title,color_title,style_title)
        i=i+1

    doc.addColumnWithStyle(coord.addRowCopy(i),values,colors=color_values,styles=style_values)
    doc.addCellWithStyle(coord.addRowCopy(i+len(values)),f"=sum({coord.addRowCopy(i).string()}:{coord.addRowCopy(i+len(values)-1).string()}",color_total,style_total)
        

## Genera totales verticales y horizontales directamente partiendo de un rango. Todos con sumas, añade un "Total" en la fila columna anterior
## @param s xlsx doc
## @param range_of_data. Range with data values
## @param keys Key of formula if List, it has all values
## @param showing When totalcolumns=True or totalrows=True only, shows a Total of Totals
def helper_totals_from_range (
                                                    doc, 
                                                    range_of_data, 
                                                    key="#SUM", 
                                                    totalcolumns=True, 
                                                    totalrows=True, 
                                                    vertical_total_title_style="BoldCenter", 
                                                    horizontal_total_title_style="BoldCenter", 
                                                    showing=False
                                                ):
    range=R.assertRange(range_of_data)
    data_rows=range.numRows()
    data_columns=range.numColumns()
    coord_horizontal_title=range.c_start.addColumnCopy(-1).addRowCopy(data_rows) 
    coord_vertical_title=range.c_start.addRowCopy(-1).addColumnCopy(data_columns)
    style_data=guess_object_style(doc.getValue(range.c_end))
    
    if totalcolumns==True and totalrows==True:
        doc.addCellWithStyle(coord_horizontal_title, _("Total"), ColorsNamed.GrayLight, horizontal_total_title_style)
        helper_totals_row(doc, coord_horizontal_title.addColumnCopy(1), [key]*data_columns,styles=style_data, row_from=range.c_start.number)
        doc.addCellWithStyle(coord_vertical_title, _("Total"), ColorsNamed.GrayLight, vertical_total_title_style)
        helper_totals_column(doc, coord_vertical_title.addRowCopy(1),[key]*(data_rows+1), styles=style_data, column_from=range.c_start.letter)
    elif totalcolumns==True:
        doc.addCellWithStyle(coord_vertical_title, _("Total"), ColorsNamed.GrayLight, vertical_total_title_style)
        helper_totals_column(doc, coord_vertical_title.addRowCopy(1),[key]*(data_rows+0), styles=style_data, column_from=range.c_start.letter)
        if showing is True:
            coord_sum_totals=coord_vertical_title.addRowCopy(data_rows+1)
            doc.addCellWithStyle(coord_sum_totals, generate_formula_total_string(key, range.c_start.addColumnCopy(data_columns+1), range.c_end.addColumnCopy(1)), ColorsNamed.GrayLight, style_data)
            doc.addCellWithStyle(coord_sum_totals.addColumnCopy(-1), _("Sum of totals"), ColorsNamed.GrayDark, style_data)
    elif totalrows==True:
        doc.addCellWithStyle(coord_horizontal_title, _("Total"), ColorsNamed.GrayLight, horizontal_total_title_style)
        helper_totals_row(doc, coord_horizontal_title.addColumnCopy(1),[key]*(data_columns+0), styles=style_data, row_from=range.c_start.number) #1 menos por la esquina
        if showing is True:
            coord_sum_totals=coord_horizontal_title.addColumnCopy(data_columns+1)
            doc.addCellWithStyle(coord_sum_totals, generate_formula_total_string(key, range.c_start.addRowCopy(data_rows+1), range.c_end.addRowCopy(1)), ColorsNamed.GrayLight, style_data)
            doc.addCellWithStyle(coord_sum_totals.addRowCopy(-1), _("Sum of totals"), ColorsNamed.GrayDark, style_data)

    return range_of_data



## Write cells from a list of ordered dictionaries
## @param lod List of ordered dictionaries
## @param keys. If None write all keys, Else must be a list of keys
## @param columns_header. Integer with the number of columns to apply color_header
## @return Range. Returns the range of the data without headers. Useful to set totals.
def helper_list_of_ordereddicts(doc, coord_start,  lod_, keys=None, columns_header=0,  color_row_header=ColorsNamed.Orange, color_column_header=ColorsNamed.Green,  color=ColorsNamed.White, styles=None):
    coord_start=C.assertCoord(coord_start)
    
    if len(lod_)==0 and keys is None:
        doc.addCellWithStyle(coord_start, _("No data to show"), ColorsNamed.Red, "BoldCenter")
        return None

        
    #Header
    if keys is None:
        keys=lod.lod_keys(lod_)
    
    for column,  key in enumerate(keys):       
        doc.addCellWithStyle(coord_start.addColumnCopy(column), key, color_row_header, "BoldCenter")
    coord_data=coord_start.addRowCopy(1)
    
    
    lor=lod.lod2lol(lod_, keys)
    
    #Generate list of colors
    colors=[]
    for i in range(len(keys)):
        if i <= columns_header-1:
            colors.append(color_column_header)
        else:
            colors.append(color)
   
    #Generate list of rows
    return doc.addListOfRowsWithStyle(coord_data, lor, colors, styles)

## Write cells from a list of ordered dictionaries
## @param lod List of ordered dictionaries
## @param keys. If None write all keys, Else must be a list of keys
## @param columns_header. Integer with the number of columns to apply color_header
## @return Range of the data
def helper_list_of_ordereddicts_with_totals(doc, coord_start,  lod, keys=None, columns_header=1,  color_row_header=ColorsNamed.Orange, color_column_header=ColorsNamed.Green,  color=ColorsNamed.White, styles=None, totalcolumns=True, totalrows=True, key="#SUM"):
    coord_start=C.assertCoord(coord_start)
    helper_list_of_ordereddicts(doc, coord_start,  lod, keys, columns_header,  color_row_header, color_column_header,  color, styles)
    range_lod=R.from_iterable_object(coord_start.addRow(1), lod)## Adds q to skip top headers
    range_lod.c_start.addColumn(columns_header) ## Adds to skip columns headers
    return helper_totals_from_range (doc, range_lod, key, totalcolumns, totalrows)
    
## It's the same of helper_list_of_ordereddicts but withth mandatory keys
## @return Range of the data
def helper_list_of_dicts(doc, coord_start,  lod, keys, columns_header=0,  color_row_header=ColorsNamed.Orange, color_column_header=ColorsNamed.Green,  color=ColorsNamed.White, styles=None):
    return helper_list_of_ordereddicts(doc, coord_start,  lod, keys, columns_header=0,  color_row_header=ColorsNamed.Orange, color_column_header=ColorsNamed.Green,  color=ColorsNamed.White, styles=None)

## Creates a new sheet called "Style names" with alll ods styles grouped by families
def helper_ods_sheet_stylenames(doc):
    doc.createSheet("Internal style names")
    doc.setColumnsWidth([5, 5])
    for column, (family,  style_names) in enumerate(doc.dict_stylenames.items()):
        doc.addCellWithStyle(C("A1").addColumn(column), family, ColorsNamed.Orange, "BoldCenter")
        doc.addColumnWithStyle(C("A2").addColumn(column), style_names)
    doc.freezeAndSelect("A2")

## This helper is used when lor length is bigger than localc limits (1048576)
## With this function you can split a lor automatically in all sheets needed
## If the number of rows is lower than max_rows makes a normal sheet
## You have to set a header of one line. 
## @param doc ODS Document
## @param sheet_name Root name of the sheet
## @param lor List of rows with data
## @param headers List of strings
## @param headers_colors Color of the sheet header
## @param columns_width None=3cm. Integer=all that value, List. Defines all columns width
## @param coord_to_freeze Coord with coord to freeze
def helper_split_big_listofrows(doc, sheet_name, lor, headers, headers_colors=ColorsNamed.Orange, columns_width=None,  coord_to_freeze="A2",  max_rows=1048575):
    ceil_=ceil(len(lor)/max_rows)
    for num_sheet in range(ceil_):
        #Sets name and headers
        if ceil_>1:
            name=_("{0} ({1} of {2})").format(sheet_name, num_sheet+1,  ceil_)
            debug(_("More than {0} rows. Spliting {1} of {2} sheets").format(max_rows, num_sheet+1,  ceil_))
        else:
            name=sheet_name
        doc.createSheet(name)
        doc.addRowWithStyle("A1", headers, headers_colors, "BoldCenter")
        
        #Sets width of columns
        if len(lor)>0:
            if columns_width is None:
                columns_width=[3]*len(lor)
            elif columns_width.__class__.__name__=="int":
                columns_width=[columns_width]*len(lor)
            doc.setColumnsWidth(columns_width)
        
        #Splits data
        from_=max_rows*num_sheet
        to_=max_rows*(num_sheet+1) if len(lor)>=max_rows*(num_sheet+1) else len(lor)
        doc.addListOfRowsWithStyle("A2", lor[from_:to_])
        doc.freezeAndSelect(C.assertCoord(coord_to_freeze))
    
    



## @package unogenerator.objects
## @brief Objects for unogenerator package
from colorama import init
from datetime import datetime


__version__ = '1.0.0'
__versiondatetime__=datetime(2024, 10, 27, 16, 11)
__versiondate__=__versiondatetime__.date()


from unogenerator.can_import_uno import can_import_uno
init()
if can_import_uno():
    try:
        from .unogenerator import ODT, ODS, ODT_Standard, ODS_Standard, LibreofficeServer
        from .commons import Coord,  Range,  ColorsNamed
    except:
        pass



## @namespace unogenerator.commons
from colorama import Fore, Style
from datetime import datetime, date, timedelta
from decimal import Decimal
from gettext import translation
from logging import info, ERROR, WARNING, INFO, DEBUG, CRITICAL, basicConfig, debug
from importlib.resources import files
from os import geteuid, remove
from subprocess import run
from tempfile import TemporaryDirectory
from uno import createUnoStruct
from pydicts.currency import Currency
from pydicts.percentage import Percentage
from pydicts import casts
from shutil import which
from socket import socket, AF_INET, SOCK_STREAM
from unogenerator import exceptions, __versiondate__

try:
    t=translation('unogenerator', files("unogenerator") / 'locale')
    _=t.gettext
except:
    _=str

class ColorsNamed:
    Black=0x111111
    White=0xFFFFFF
    Blue=0x9999ff
    Red=0xFF9999
    Green=0xc0FFc0
    Orange=0xffdca8
    Yellow=0xffffc0
    GrayLight=0xd2ced4
    GrayDark=0xa29ea4
    GrayVeryDark=0x726e74

def datetime2uno( dt):
    r=createUnoStruct("com.sun.star.util.DateTime")
    r.Year=dt.year
    r.Month=dt.month
    r.Day=dt.day
    r.Hours=dt.hour
    r.Minutes=dt.minute
    r.Seconds=dt.second
    return r


def is_port_opened(host, port):
    sock = socket(AF_INET, SOCK_STREAM)
    result = sock.connect_ex((host,port))
    sock.close()
    if result == 0:
       return False
    else:
       return True



## Converts a com.sun.star.util.DateTime to datetime
def uno2datetime(r):
    try:
        return datetime(r.Year, r.Month, r.Day, r.Hours, r.Minutes, r.Seconds)
    except:
        debug(_("There was a problem converting com.sun.star.util.DateTime to datetime."))

## Function used in argparse_epilog
## @return String
def argparse_epilog():
    return _("Developed by Mariano Muñoz 2020-{}").format(__versiondate__.year)


## Allows to operate with columns letter names
## @param letter String with the column name. For example A or AA...
## @param number Columns to move
## @return String With the name of the column after movement
def columnAdd(letter, number):
    letter_value=column2number(letter)+number
    return number2column(letter_value)


def rowAdd(letter,number):
    return str(int(letter)+number)

## Convierte un número  con el numero de columna al nombre de la columna de hoja de datos
##
## Number to Excel-style column name, e.g., 1 = A, 26 = Z, 27 = AA, 703 = AAA.
def number2column(n):
    name = ''
    while n > 0:
        n, r = divmod (n - 1, 26)
        name = chr(r + ord('A')) + name
    return name

## Convierte una columna de hoja de datos a un número
##
## Excel-style column name to number, e.g., A = 1, Z = 26, AA = 27, AAA = 703.
def column2number(name):
    n = 0
    for c in name:
        n = n * 26 + 1 + ord(c) - ord('A')
    return n

## Converts a column name to a index position (number of column -1)
def column2index(name):
    return column2number(name)-1

## Convierte el nombre de la fila de la hoja de datos a un índice, es decir el número de la fila -1
def row2index(number):
    return int(number)-1

## Covierte el nombre de la fila de la hoja de datos a un  numero entero que corresponde con el numero de la fila
def row2number(strnumber):
    return int(strnumber)

## Convierte el numero de la fila al nombre de la fila en la hoja de datos , que corresponde con un string del numero de la fila
def number2row(number):
    return str(number)
    
## Convierte el indice de la fila al numero cadena de la hoja de datos
def index2row(index):
    return str(index+1)
    
## Convierte el indice de la columna a la cadena de letras de la columna de la hoja de datos
def index2column(index):
    return number2column(index+1)

class Coord:
    """
        Class that manage spreadsheet coordinates (letter + number)
    """
    def __init__(self, value):
        """
            Coord constructor
            
            Value can be:
            - String
        """
        self.letter=None
        self.number=None
        if value.__class__==str:
            self.letter, self.number=self.__extract(value)
        else:
            raise exceptions.CoordException(_("Coord constructor must have a str or a Coord parameter and is a {0}").format(value.__class__))
            
        
    @staticmethod
    def assertCoord(o):
        if o.__class__==Coord:
            return o
        elif o.__class__==str:
            return Coord(o)
        raise exceptions.CoordException(_("I can't convert {} (class: {}) to a Coord").format(o, o.__class__))
        
    ## Creates a Coord object from spreadsheet index coords
    @classmethod
    def from_index(cls, column_index, row_index):
        return cls(index2column(column_index)+index2row(row_index))
        
            
    ## Creates a Coord object from spreadsheet letters
    @classmethod
    def from_letters(cls, column, letter):
        return cls(column+letter)

    def __repr__(self):
        return f"Coord <{self.string()}>"
        
    def __str__(self):
        return self.string()

    def __eq__(self, b):
        b=Coord.assertCoord(b)
        if self.letter==b.letter and self.number==b.number:
            return True
        return False
        
    def __extract(self, strcoord):
        allowed="ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890"
        letter=""
        number=""
        for i, l in enumerate(strcoord):
            if not l in allowed:
                raise exceptions.CoordException(_("Coord can have this character: {0}").format(l))
            
            if l.isdigit()==False:
                letter=letter+l
            else:
                number=strcoord[i:]
                break
                
        if len(letter)==0:
            raise exceptions.CoordException(_("Coord letter can't be empty"))
                
        try:
            number=str(int(number))
        except ValueError:
            raise exceptions.CoordException(_("Coord couldn't be created with '{0}'").format(strcoord))
        
        return (letter,number)

    ## Returns Coord string
    ## @return string For example "Z1"
    def string(self):
        return self.letter+self.number

    ## Add a number of rows to the Coord and return itself
    ## @param num Integer Can be positive and negative. When num is negative, if Coord.number is less than 1, returns 1
    def addRow(self, num=1):
        if self.numberIndex()+num<0:
            self.number="1"
        else:
            self.number=rowAdd(self.number, num)
        return self

    ## Add a number of columns/letters to the Coord and return itself
    ## @param num Integer. Can be positive and negative. When num is negative, if Coord.letter is less than A, returns A.
    def addColumn(self, num=1):
        if self.letterIndex()+num<0:
            self.letter="A"
        else:
            self.letter=columnAdd(self.letter, num)
        return self

    ## Add a number of rows to the Coord and return a copy of the object
    ## @param num Integer Can be positive and negative. When num is negative, if Coord.number is less than 1, returns 1
    def addRowCopy(self, num=1):
        r=Coord(self.string())
        r.addRow(num)
        return r

    ## Add a number of columns/letters to the Coord and return a copy of the objject
    ## @param num Integer. Can be positive and negative. When num is negative, if Coord.letter is less than A, returns A.
    def addColumnCopy(self, num=1):
        r=Coord(self.string())
        r.addColumn(num)
        return r

    ## Reterns the letter/column index (letterPosition()-1)
    ## @return int
    def letterIndex(self):
        return column2index(self.letter)

    ## Returns the letter/column position
    ## @return int 
    def letterPosition(self):
        return column2number(self.letter)

    ## Returns the number/row index (numberPosition()-1)
    def numberIndex(self):
        return row2index(self.number)

    ## Returns the number/row position
    def numberPosition(self):
        return row2number(self.number)


## Class that manages spreadsheet Ranges for ods and xlsx
class Range:
    def __init__(self,strrange):
        self.c_start, self.c_end=self.__extract(strrange)

    def __eq__(self, b):
        b=Range.assertRange(b)
        if self.c_start==b.c_start and self.c_end==b.c_end:
            return True
        return False
        
    def __repr__(self):
        return f"Range <{self}>"
        
    def __str__(self):
        return self.string()
        
    @classmethod
    def from_uno_range(cls, uno_range):
        return cls.from_coords_indexes(uno_range.RangeAddress.StartColumn, uno_range.RangeAddress.StartRow, uno_range.RangeAddress.EndColumn, uno_range.RangeAddress.EndRow)
    
    @classmethod
    def from_coords_indexes(cls, start_letter_index, start_number_index, end_letter_index, end_number_index):
        c_start=Coord.from_index(start_letter_index, start_number_index)
        c_end=Coord.from_index(end_letter_index, end_number_index)
        return cls(f"{c_start}:{c_end}")

    ## Creates a Range object from itself stard and c_end coords
    @classmethod
    def from_coords(cls, c_start, c_end):
        c_start=Coord.assertCoord(c_start)
        c_end=Coord.assertCoord(c_end)
        return cls(f"{c_start}:{c_end}")
    
    ## Creates a Range object from Changing the start_coord of a range
    @classmethod
    def from_start_coord_change(cls, range, new_start_coord):
        return cls.from_columns_rows(new_start_coord, range.numColumns(), range.numRows())
        
    ## Creates a Range object from a coord_start and the number of columms and rows of therange
    @classmethod
    def from_columns_rows(cls, coord_start, number_columns,  number_rows):
        c_start=Coord.assertCoord(coord_start)
        c_end=c_start.addRowCopy(number_rows-1).addColumnCopy(number_columns-1)
        return cls(f"{c_start}:{c_end}")

    @classmethod
    def from_iterable_object(cls,  coord_start, o):
        coord_start=Coord.assertCoord(coord_start)
        r=Range(f"{coord_start.string()}:{coord_start.string()}")

        if len(o)==0:
            return r

        if o[0].__class__.__name__ in ("list","dict", "OrderedDict"): #Iterables con len
            len_rows=len(o)-1
            len_columns=len(o[0])-1
        else:
            len_rows=0
            len_columns=len(o)-1
        return Range(f"{coord_start.string()}:{coord_start.addRowCopy(len_rows).addColumnCopy(len_columns).string()}")

    ##Return the outcome of the test b in a. Note the reversed operands.
    def __contains__(self, b):
        if (    b.letterIndex()>=self.c_start.letterIndex() and 
                b.letterIndex()<=self.c_end.letterIndex() and 
                b.numberIndex()>=self.c_start.numberIndex() and 
                b.numberIndex()<=self.c_end.numberIndex())==True:
            return True
        return False
        
    def uno_range(self, sheet):
        """
            Returns the uno range of the current sheet
            @param sheet is current sheet. ODS.sheet parameter
        """
        range_indexes=[self.c_start.letterIndex(), self.c_start.numberIndex(), self.c_end.letterIndex(), self.c_end.numberIndex()]
        return sheet.getCellRangeByPosition(*range_indexes)

    ## Converts a string to a Range. Returns None if conversion can't be done
    def __extract(self,strrange):
        if not strrange.__class__==str:
            raise exceptions.RangeException(_("Range constructor must receive a string but got: {0} ({1})").format(strrange, strrange.__class__))
        
        a=strrange.split(":")
        if not len(a)==2:
            raise exceptions.RangeException(_("This is not a range: {0}").format(strrange))
            
        try:
            c_start=Coord(a[0])
        except exceptions.CoordException:
            raise exceptions.RangeException(_("Range start coord is wrong: {0}").format(a[0]))
        try:
            c_end=Coord(a[1])
        except exceptions.CoordException:
            raise exceptions.RangeException(_("Range end coord is wrong: {0}").format(a[1]))
        return (c_start, c_end)

    ## String of a range in spreadsheets
    def string(self):
        return "{}:{}".format(self.c_start.string(), self.c_end.string())

    ## Number of range of the range
    def numRows(self):
        return row2number(self.c_end.number)-row2number(self.c_start.number) +1

    ## Number of columns of the range
    def numColumns(self):
        return column2number(self.c_end.letter)-column2number(self.c_start.letter) +1

    ## Returns a Range object even o is a str or a Range
    @staticmethod
    def assertRange(o):
        if o.__class__==Range:
            return o
        elif o.__class__==str:
            return Range(o)
        raise exceptions.RangeException(_("I can't convert {} (class: {}) to a Range").format(o, o.__class__))


    ## Adds a row to the c_end Coord, so it adds a row to the range
    def addRowAfter(self, num=1):
        self.c_end=self.c_end.addRow(num)
        return self

    def addRowAfterCopy(self, num=1):
        r=Range(self.string())
        r.addRowAfter(num)
        return r

    ## Adds a column to the c_end Coord, so it adds a column to the range
    def addColumnAfter(self, num=1):
        self.c_end=self.c_end.addColumn(num)
        return self

    def addColumnAfterCopy(self, num=1):
        r=Range(self.string())
        r.addColumnAfter(num)
        return r

    ## Adds a row to the top of the c_start Coord, so it adds a row to the range. If c_start Coord number is 1 returns the same Coord
    def addRowBefore(self, num=1):
        self.c_start=self.c_start.addRow(-num)
        return self

    def addRowBeforeCopy(self, num=1):
        r=Range(self.string())
        r.addRowBefore(num)
        return r

    ## Adds a column to the c_end Coord, so it adds a column to the range. If c_start Coord letter is A, returns the same Coord
    def addColumnBefore(self, num=1):
        self.c_start=self.c_start.addColumn(-num)
        return self

    def addColumnBeforeCopy(self, num=1):
        r=Range(self.string())
        r.addColumnBefore(num)
        return r

    ## Returns a list of tuples(column_index, row_index)
    def indexes_list(self, plain=False):
        r=[]
        if plain is True:
            for letter_index in range(self.c_start.letterIndex(), self.c_end.letterIndex()+1):
                for number_index in range(self.c_start.numberIndex(), self.c_end.numberIndex()+1):
                    r.append((letter_index, number_index))
        else:
            for number_index in range(self.c_start.numberIndex(), self.c_end.numberIndex()+1):
                row=[]
                for letter_index in range(self.c_start.letterIndex(), self.c_end.letterIndex()+1 ):
                    row.append((letter_index, number_index))
                r.append(row)
        return r
        
    ## Returns a list of rows of all Coord objects in the range
    def coords_list(self, plain=False):
        r=[]
        if plain is True:
            for letter_index, number_index in self.indexes_list(plain):
                r.append(Coord.from_index(letter_index, number_index))
        else:
            for row  in self.indexes_list(plain):
                r2=[]
                for  letter_index , number_index  in row:
                    r2.append(Coord.from_index(letter_index, number_index))
                r.append(r2)
        return r

## Sets debug sustem, needs
## @param args It's the result of a argparse     args=parser.parse_args()        
def addDebugSystem(level):
    logFormat = "%(asctime)s.%(msecs)03d %(levelname)s %(message)s [%(module)s:%(lineno)d]"
    dateFormat='%F %I:%M:%S'

    if level=="DEBUG":#Show detailed information that can help with program diagnosis and troubleshooting. CODE MARKS
        basicConfig(level=DEBUG, format=logFormat, datefmt=dateFormat)
    elif level=="INFO":#Everything is running as expected without any problem. TIME BENCHMARCKS
        basicConfig(level=INFO, format=logFormat, datefmt=dateFormat)
    elif level=="WARNING":#The program continues running, but something unexpected happened, which may lead to some problem down the road. THINGS TO DO
        basicConfig(level=WARNING, format=logFormat, datefmt=dateFormat)
    elif level=="ERROR":#The program fails to perform a certain function due to a bug.  SOMETHING BAD LOGIC
        basicConfig(level=ERROR, format=logFormat, datefmt=dateFormat)
    elif level=="CRITICAL":#The program encounters a serious error and may stop running. ERRORS
        basicConfig(level=CRITICAL, format=logFormat, datefmt=dateFormat)
    info("Debug level set to {}".format(level))





def generate_formula_total_string(key, coord_from, coord_to):
    if key == "#SUM":
        s="=SUM({}:{})".format(coord_from.string(), coord_to.string())
    elif key == "#AVG":
        s="=AVERAGE({}:{})".format(coord_from.string(), coord_to.string())
    elif key == "#MEDIAN":
        s="=MEDIAN({}:{})".format(coord_from.string(), coord_to.string())
    else:
        s=key
    return s

def guess_object_style(o):
    if o is None:
        return "Normal"
    elif o.__class__.__name__=="int":
        return "Integer"
    elif o.__class__.__name__=="timedelta":
        return "TimedeltaSeconds"#TimedeltaISO exits but you can't add or supr
    elif o.__class__.__name__=="str":
        return "Normal"
    elif o.__class__.__name__ in ["Currency", "Money" ]:
        return o.currency
    elif o.__class__.__name__=="Percentage":
        return "Percentage"
    elif o.__class__.__name__ in ("Decimal", "float"):
        return  "Float2"
    elif o.__class__.__name__=="datetime":
        return "Datetime"
    elif o.__class__.__name__=="date":
        return "Date"
    elif o.__class__.__name__=="time":
        return "Time"
    elif o.__class__.__name__=="bool":
        return "Bool"
    else:
        info("guess_object_style not guessed {}".format( o.__class__))
        return "Bold"
        
def datetime2localc1989(o):
    """
        Converts a datetime to a localc1989 values. Used with getValues with getDataArray method
        datetime(2023,12,2,12,0,0) ==> 45262.5
    """
    delta = o -  datetime(1899, 12, 30)
    return float(delta.days) + float(delta.seconds) / 86400
    
def localc19892datetime(value):
    return  datetime(1899, 12, 30)+timedelta(days=int(value),  seconds=(value-int(value))*86400)
    
    
def date2localc1989(o):
    """
        Converts a date to a localc1989 values. Used with getValues with getDataArray method
        date(2023,12,2) ==> 45262.0
    """
    delta = o -  date(1899, 12, 30)
    return float(delta.days) 

def localc19892date(o):
    return date(1899, 12, 30) + timedelta(days=o)

def time2localc1989(o):
    seconds=o.hour*3600+o.minute*60+o.second
    return float(seconds) / 86400
    
def localc19892time(o):
    ##If you need to get datetime.time value, you can use this trick:
    ##my_time = (datetime(1970,1,1) + timedelta(seconds=my_seconds)).time()
    ##You cannot add timedelta to time, but can add it to datetime.
    return (datetime(1970,1,1) + timedelta(seconds=o*86400)).time()



    
## Used to change port when there are multiple sockets of LibreOffice accepting
def next_port(last,  first_port,  instances):
    if last==first_port+instances -1:
        return first_port
    else:
        return last+1

    
        
## Converts an string or a float to an object.
## Used to cast getDataArray
## string_float. String or float or None
## cast "int", "float, ....
## If there is an error, returns the value
def string_float2object(string_float, cast):
    if string_float is None:
        return None
    if cast=="int":
        try:
            return int(string_float)
        except:
            return string_float
    elif cast=="float":
        try:
            return float(string_float)
        except:
            return string_float
    elif cast=="Decimal":
        try:
            return Decimal(str(string_float))
        except:
            return string_float
    elif cast=="datetime":
        try:
            return localc19892datetime(string_float)
        except:
            return string_float
    elif cast=="date":
        try:
            return localc19892date(string_float)
        except:
            return string_float
    elif cast=="time":
        try:
            return localc19892time(string_float)
        except:
            return string_float
    elif cast=="timedelta":
        if string_float.__class__== str:
            return casts.str2timedelta(string_float)
        else:
            return timedelta(seconds=string_float)
        
    elif cast=="bool":
        if string_float==0:
            return False
        elif string_float==1:
            return True
        else:
            return string_float
    elif cast in ["EUR", "€"]:
        try:
            return Currency(string_float, "EUR")
        except:
            return string_float
    elif cast in ["USD", "$"]:
        try:
            return Currency(string_float, "USD")
        except:
            return string_float
    elif cast=="Percentage":
        if string_float.__class__.__name__=="str":
            return string_float
        try:
            return Percentage(string_float, 1)
        except:
            return string_float
    return string_float

def is_root():
    return geteuid() == 0
   
def is_formula(s):
    if s is None:
        return False
    s=str(s)
    if s.startswith("=") or s.startswith("+"):
        return True
    return False
   
def red(s):
    return Style.BRIGHT + Fore.RED + str(s) + Style.RESET_ALL
def green(s):
    return Style.BRIGHT + Fore.GREEN + str(s) + Style.RESET_ALL
def magenta(s):
    return Style.BRIGHT + Fore.MAGENTA + str(s) + Style.RESET_ALL
def yellow(s):
    return Style.BRIGHT + Fore.YELLOW + str(s) + Style.RESET_ALL
def white(s):
    return Style.BRIGHT + str(s) + Style.RESET_ALL
    
## Removes border white space from a path or a byte array of a image. It uses convert from imagemagick
## @param type Image extension
def bytes_after_trim_image(filename_or_bytessequence, type):
    with TemporaryDirectory() as tmpdirname:
        filename_trimed=f"{tmpdirname}/filename_trimed.{type}"
        #Converts bytes to temporary file
        if filename_or_bytessequence.__class__.__name__=="bytes":
            filename_to_trim=f"{tmpdirname}/filename_to_trim.{type}"
            with open(filename_to_trim, "w+b") as f:
                f.write(filename_or_bytessequence)
        else:
            filename_to_trim=filename_or_bytessequence

        #Trims white space
        if which("magick") is not None:
            p=run(f"magick '{filename_to_trim}' -trim +repage '{filename_trimed}'",  shell=True)
            if p.returncode==0:
                #Alter bytes if converted
                with open(filename_trimed, "r+b") as f:
                    bytes_=f.read()
                    return bytes_
        elif which("convert") is not None:
            p=run(f"convert -trim +repage '{filename_to_trim}' '{filename_trimed}'",  shell=True)
            if p.returncode==0:
                #Alter bytes if converted
                with open(filename_trimed, "r+b") as f:
                    bytes_=f.read()
                    return bytes_
        else:
            print(_("There was an error triming image. Is convert command (from Imagemagick) installed?"))


def are_all_values_of_the_list_the_same(list_):
    """
        Returns true if all values of the list are the same
        Devuelve True si la lista es vacía
    """
    if len(list_)==0:
        return True
        
    value=list_[0]
    for i in range(1, len(list_)):
        if not list_[i]==value:
            return False
    return True
    
def remove_without_errors(filename): 
    try: 
        remove(filename) 
    except OSError as e: 
        print(_("Error deleting: {0} -> {1}").format(filename, e.strerror)) 



## @namespace unogenerator.unogenerator
## @brief Package that allows to read and write Libreoffice ods and odt files

from datetime import datetime
from os import path, makedirs
from uno import getComponentContext, createUnoStruct, systemPathToFileUrl, Any, ByteSequence
from com.sun.star.beans import PropertyValue
from com.sun.star.text import ControlCharacter
from com.sun.star.awt import Size
from com.sun.star.sheet.ConditionEntryType import COLORSCALE
from com.sun.star.style.ParagraphAdjust import RIGHT,  LEFT
from com.sun.star.style.BreakType import PAGE_BEFORE, PAGE_AFTER
from gettext import translation
from logging import warning, debug
from importlib.resources import files
from os import system
from pydicts import lol, casts
from shutil import copyfile
from socket import socket, AF_INET, SOCK_STREAM
from subprocess import Popen, PIPE
from tempfile import TemporaryDirectory
from time import sleep
from unogenerator import __version__, exceptions
from unogenerator.commons import Coord, ColorsNamed,  Range as R, datetime2uno, guess_object_style, datetime2localc1989, date2localc1989,  time2localc1989,  is_formula, uno2datetime, string_float2object
from pydicts.currency import Currency
from pydicts.percentage import Percentage

def createUnoService(serviceName):
#        resolver = localContext.ServiceManager.createInstance('com.sun.star.bridge.UnoUrlResolver')
  return getComponentContext().ServiceManager.createInstance(serviceName)
  
try:
    t=translation('unogenerator', files("unogenerator") / 'locale')
    _=t.gettext
except:
    _=str

class LibreofficeServer:
    def __init__(self):
        self.pid=None
        self.start()
        
    ## This method allows to use with statement. 
    ## with LibreofficeServer() as server:
    ##      with ODS_Standard
    ## First calls __init__ with None and 2002, then enter, then exit
    def __enter__(self):
        return self
        
    ## Exit function to use with with statement. __enter__ defines enter in with
    def __exit__(self, *args, **kwargs):
        self.stop()

    def start(self):
        # Gets an unued port
        with socket(AF_INET, SOCK_STREAM) as s:
            s.bind(('', 0))  # Bind to port 0 to let the OS assign a free port
            self.port=s.getsockname()[1]

        command=f'loffice --accept="socket,host=localhost,port={self.port};urp;StarOffice.ServiceManager" -env:UserInstallation=file:///tmp/unogenerator{self.port} --headless  --nologo  --norestore'
        process=Popen(command, stdout=PIPE, stderr=PIPE, shell=True)       
        self.pid=process.pid
        
    def stop(self):
        system(f'pkill -f socket,host=localhost,port={self.port};urp;StarOffice.ServiceManager')
        system(f'rm -Rf /tmp/unogenerator{self.port}')

class ODF:
    def __init__(self, template=None,  server=None):
        """
            Common class for ODF instances

            @param template path to template
            @type string
            @param server Server object to use
            @type LibreofficeServer
        """        
        self.start=datetime.now()
        self.server=LibreofficeServer() if server is None else server #Assigns server or auto launch if None
        self.autoserver=server==None
        self.template=None if template is None else systemPathToFileUrl(path.abspath(template))
        maxtries=300
        
        for i in range(maxtries):
            try:
                localContext = getComponentContext()
                resolver = localContext.ServiceManager.createInstance('com.sun.star.bridge.UnoUrlResolver')
                ## self.ctx parece que es mi contexto para servicios
                self.ctx = resolver.resolve(f'uno:socket,host=127.0.0.1,port={self.server.port};urp;StarOffice.ComponentContext')
                self.desktop = self.ctx.ServiceManager.createInstance('com.sun.star.frame.Desktop')
                self.graphicsprovider=self.ctx.ServiceManager.createInstance("com.sun.star.graphic.GraphicProvider")                   
                args=(
                    PropertyValue('AsTemplate',0,True,0),
                )
                if self.__class__  in (ODS, ODS_Standard):
                    if self.template is None:
                        self.document=self.desktop.loadComponentFromURL('private:factory/scalc','_blank',8,())
                    else:
                        self.document=self.desktop.loadComponentFromURL(self.template,'_blank', 8, args)
                    self.sheet=self.setActiveSheet(0)
                else: #ODT
                    if self.template is None:
                        self.document=self.desktop.loadComponentFromURL('private:factory/swriter','_blank', 8, ())
                    else:
                        self.document=self.desktop.loadComponentFromURL(self.template,'_blank', 8, args)
                    self.cursor=self.document.Text.createTextCursor()
                self.dict_stylenames=self.dictionary_of_stylenames()
                break
            except Exception as e:
                sleeptime=0.25
                sleep(sleeptime)
                if i==maxtries - 1:
                    print(_("This process died after trying to connect to port {0} during {1} seconds. Error: {2}").format(self.loserver_port, maxtries*sleeptime, e))

    ## This method allows to use with statement. 
    ## with ODS() as doc:
    ##      doc.createSheet("WITH")
    ## First calls __init__ with None and 2002, then enter, then exit
    def __enter__(self):
        return self
        
    ## Exit function to use with with statement. __enter__ defines enter in with
    def __exit__(self, *args, **kwargs):
        self.close()

    def calculateAll(self):
        self.document.calculateAll()

    def close(self):
        try:
            self.document.dispose()
        except:
            print (_("Error closing ODF instance"))
        finally:
            if self.autoserver is True:
                self.server.stop()

    ## Generate a dictionary_of_styles with families as key, and a list of string styles as value
    def dictionary_of_stylenames(self):
        stylefam=self.document.StyleFamilies
        d={}
        for sf_index,  sf in enumerate(stylefam):
            r=[]
            styles=list(sf.getElementNames())
            styles.sort()
            for style in styles:
                r.append(style)
            d[stylefam.getElementNames()[sf_index]]=r
        return d

    # Returns a list of page styles objects
    def getPageStylesObjects(self):
        r=[]
        family=self.document.StyleFamilies.getByName('PageStyles')
        for i in range(family.Count):
            r.append(family.getByIndex(i))
        return r
        

    def setLanguage(self, language, country):
        self.language="es"
        self.country="ES"

    def getMetadata(self):
        print("Author",  self.document.DocumentProperties.Author)
        print(dir(self.document.DocumentProperties))
        return {
            "Author": self.document.DocumentProperties.Author, 
            "Description": self.document.DocumentProperties.Description, 
            "Subject": self.document.DocumentProperties.Subject, 
            "Keywords": self.document.DocumentProperties.Keywords, 
            "CreationDate": uno2datetime(self.document.DocumentProperties.CreationDate), 
            "ModificationDate": uno2datetime(self.document.DocumentProperties.ModificationDate), 
            "Title": self.document.DocumentProperties.Title, 
        }
        
    ##Only sets a value when it's different of [] or ""
    def setMetadata(self, title="",  subject="", author="", description="", keywords=[], creationdate=datetime.now()):
        try:
            if author!="":
                self.document.DocumentProperties.Author=author
            self.document.DocumentProperties.Generator=f"UnoGenerator-{__version__}"
            if description!="":
                self.document.DocumentProperties.Description=description
            if subject!="":
                self.document.DocumentProperties.Subject=subject
            if keywords!="":
                self.document.DocumentProperties.Keywords=keywords
            self.document.DocumentProperties.CreationDate=datetime2uno(creationdate)
            self.document.DocumentProperties.ModificationDate=datetime2uno(creationdate)
            if title!="":
                self.document.DocumentProperties.Title=title
        except:
            print("Error setting metadata. Sometimes fails with concurrent process")

    def deleteAll(self):
        self.executeDispatch(".uno:SelectAll")
        self.executeDispatch(".uno:Delete")
        
    def executeDispatch(self, command):
        oDisp = createUnoService("com.sun.star.frame.DispatchHelper")
        oDisp.executeDispatch(self.document.getCurrentController().Frame, command, "", 0, ())
        
    ## Poner el tray en el resover y cambiar el puerto cuando except
    
    def loadStylesFromFile(self, filename, overwrite=True):
        styleoptions=list(self.document.StyleFamilies.StyleLoaderOptions)#it's a tuple
        styleoptions.pop()
        styleoptions.append(PropertyValue("OverwriteStyles",0,overwrite,0))
        self.document.StyleFamilies.loadStylesFromURL(systemPathToFileUrl(filename), tuple(styleoptions))

        
                   
class ODT(ODF):
    def __init__(self, template=None, server=None):
        ODF.__init__(self, template, server)

    def save(self, filename, overwrite_template=False):
        if filename==self.template and overwrite_template is False:
            print(_("You can't use the same filename as your template or you will overwrite it."))
            print(_("You can force it, setting overwrite_template paramter to True"))
            print(_("Document hasn't been saved."))
            return
        if filename.endswith(".odt") is False:
            print(_("Filename extension must be 'odt'."))
            print(_("Document hasn't been saved."))
            return
        ## SAVE FILE
        args=(
            PropertyValue('FilterName',0,'writer8',0),
            PropertyValue('Overwrite',0,True,0),
        )
    

        with TemporaryDirectory() as tmpdirname:
                tempfile=f"{tmpdirname}/{path.basename(filename)}"
                self.document.storeAsURL(systemPathToFileUrl(tempfile), args)
                makedirs(path.dirname(path.abspath(filename)), exist_ok=True)
                copyfile(tempfile, filename)
        
                
    def export_pdf(self, filename):
        if filename.endswith(".pdf") is False:
            print(_("Filename extension must be 'pdf'."))
            print(_("Document hasn't been saved."))
            return
        args=(
            PropertyValue('FilterName',0,'writer_pdf_Export',0),
            PropertyValue('Overwrite',0,True,0),
        )
        with TemporaryDirectory() as tmpdirname:
            tempfile=f"{tmpdirname}/{path.basename(filename)}"
            self.document.storeToURL(systemPathToFileUrl(tempfile), args)
            
            makedirs(path.dirname(path.abspath(filename)), exist_ok=True)
            copyfile(tempfile, filename)
         
    def export_docx(self, filename):
        if filename.endswith(".docx") is False:
            print(_("Filename extension must be 'docx'."))
            print(_("Document hasn't been saved."))
            return
        args=(
            PropertyValue('FilterName',0,'MS Word 2007 XML',0),
            PropertyValue('Overwrite',0,True,0),
        )

        with TemporaryDirectory() as tmpdirname:
            tempfile=f"{tmpdirname}/{path.basename(filename)}"
            self.document.storeToURL(systemPathToFileUrl(tempfile), args)
            makedirs(path.dirname(path.abspath(filename)), exist_ok=True)
            copyfile(tempfile, filename)

    ## This method has problems with ().\ *// especial characters. Findall and replace works fine
    ## Last is the descriptor returned after first found
    ## @return None if  not found. Descriptor(can be used again with last) of the last found
    def find_and_replace(self, find, replace="", last=None, log=False):
        search=self.document.createSearchDescriptor()
        search.SearchString=find
        if last is None:
            found=self.document.findFirst(search)
        else:
            found=self.document.findNext(last,search)
        
        if found is not None:
            if found.HyperLinkTarget != "" or found.HyperLinkName != "" or found.HyperLinkURL != "":
                url=found.HyperLinkURL
                target=found.HyperLinkTarget
                found.setString(replace)
                found.HyperLinkURL=url
                found.HyperLinkTarget=target
            else:
                found.setString(replace)
            self.cursor=found
        else:
            if log is True:
                warning(f"'{find}' was not found in the document'")
        return found

    def findall_and_replace(self, find, replace="", log=False):
        search=self.document.createReplaceDescriptor()
        search.setSearchString(find)
        search.setReplaceString(replace)
        found=self.document.findFirst(search)
        if found is not None:
            number=self.document.replaceAll(search)
            debug(f"Replaced {number} times '{find}' by '{replace}'")
            self.cursor=found
            return number
        else:
            if log is True:
                debug(f"'{find}' was not found in the document'")
            return 0

    ## Sets the cursor after found string
    def find(self, find, log=False):
        self.first_and_replace(find, find, log)

    ## Find a string and deleteAll untill the end of the document
    ## Usefull to delete document part with styles in templates
    def find_and_delete_until_the_end_of_document(self, find):
        found=self.find_and_replace(find, "", None, False)
        if found is not None:
#            found.setString("")
            self.cursor=found
            
            oVC = self.document.getCurrentController().getViewCursor()#		'Create View Cursor oVC at the end of document by default?
            oVC.gotoEnd(False)
            self.cursor.gotoRange(oVC,True)			#'Move Text Cursor to same location as oVC while selecting text in between (True)
            self.cursor.setString("")
        else:
            warning(f"'{find}' was not found in the document'")
        
    def addString(self, text, style=None, paragraphBreak=False):
        if style is not None:
            self.cursor.setPropertyValue("ParaStyleName", style)
        self.document.Text.insertString(self.cursor, text, False)
        if paragraphBreak is True:
            self.document.Text.insertControlCharacter(self.cursor, ControlCharacter.PARAGRAPH_BREAK, False)
        
        
        
    def addStringHyperlink(self,  name,  url, paragraphBreak=False):

        oVC = self.document.getCurrentController().getViewCursor()#		'Create View Cursor oVC at the end of document by default?
        text=oVC.getText()
        text.insertString(oVC, name,  True)
        oVC.HyperLinkTarget=url
        oVC.HyperLinkURL=url
#        
        oVC.goRight(len(name), False)
        self.cursor.gotoRange(oVC,False)			#'Move Text Cursor to same location as oVC while selecting text in between (True)
#        self.document.Text.insertString(self.cursor, " ",  False)

        if paragraphBreak is True:
            self.document.Text.insertControlCharacter(self.cursor, ControlCharacter.PARAGRAPH_BREAK, False)
        
    def addParagraph(self,  text,  style="Standard"):
        self.cursor.setPropertyValue("ParaStyleName", style)
        self.document.Text.insertString(self.cursor, text, False)
        self.document.Text.insertControlCharacter(self.cursor, ControlCharacter.PARAGRAPH_BREAK, False)
        
    ## @param list Pueden ser str, o textcontent objects
    def addParagraphComplex(self,  list,  style):
        self.cursor.setPropertyValue("ParaStyleName", style)
        for o in list:
            if o.__class__==str:
                self.document.Text.insertString(self.cursor, o, False)
            elif o.ImplementationName=="SwXTextGraphicObject":## Es un pyuno object raro no mostraba bvien
                self.document.Text.insertTextContent(self.cursor, o, False)               
        self.document.Text.insertControlCharacter(self.cursor, ControlCharacter.PARAGRAPH_BREAK, False)
        
        
    def addHTMLBlock(self, htmlcode):  
        oStream = self.ctx.ServiceManager.createInstanceWithContext("com.sun.star.io.SequenceInputStream", self.ctx)
        oStream.initialize((ByteSequence(htmlcode.encode()),))

        prop1 = PropertyValue()
        prop1.Name  = "FilterName"
        prop1.Value = "HTML (StarWriter)"
        prop2 = PropertyValue()
        prop2.Name = "InputStream" 
        prop2.Value = oStream
        oVC = self.document.getCurrentController().getViewCursor()#		'Create View Cursor oVC at the end of document by default?
        self.cursor.insertDocumentFromURL("private:stream", (prop1, prop2))
        oVC.goRight(len(htmlcode), False)
        self.cursor.gotoRange(oVC,False)		

        
    def pixelsto100thmm(self, pixels):
        ##                #OOo uses, 1440 twips per inch, which equals 56.7 twips per mm,
        ##The functions TwipsPerPixelX() and TwipsPerPixelY() (both of which return 15 on my system),
        ##Shape is in 100ths of a mm,
        ##Therefore,
        ##PixelsWidth = (1/15) * 56.7 * (ShapeSize.Width/100) [ (pixels/twips) * (twips/mm) * ((mm*100)/100) = pixels ]
        ##PixelsWidth = 0.0378 * ShapeSize.Width
        cm=pixels*15*1/56.7*100#pixels* [twips/pixels]*[mm/twips] * [*100thmm/1mm]
        #print("Converting pixels ",  pixels, "to thmm", cm)

        return cm  
            
        
    ## Returns a text content that can be inserted with document.Text.insertTextContent(cursor,image, False)
    
    ## @param filename_or_bytessequence, Can be a filename path or a bytesequence
    ## @param anchortype AS_CHARACTER, AT_PARAGRAPH
    ## @param name None if we want to use lo default name
    ## @param width 
    ## @param height Puede ser None or a value. If a value, witdth will be set  automatically
    ##      -Width:None -Heigth:None. Usa tamaño por defecto de la imagen (Opción por defecto)
    ##      - Width: 5  -Height:None. Pone 5cm de ancho y la altura automática
    ##      - Width: None  -Height:None5. Pone la anchura automática y 5cm de alto
    ##      - Width: 5  -Height:5. Pone 5cm de ancho de alto
    
    def textcontentImage(self, filename_or_bytessequence, width=None,  height=None, anchortype="AS_CHARACTER", name=None, linked=False ):
        if filename_or_bytessequence.__class__.__name__=="bytes":
            bytes_stream = self.ctx.ServiceManager.createInstanceWithContext('com.sun.star.io.SequenceInputStream', self.ctx)
            bytes_stream.initialize((ByteSequence(filename_or_bytessequence),)) ##Método para inicializar un servicio
            oProps=(
                PropertyValue('InputStream',0, bytes_stream,0),
            )        
            
        else:
            oProps=(
                PropertyValue('URL',0,systemPathToFileUrl(str(filename_or_bytessequence)),0),
                PropertyValue('LoadAsLink',0, linked,0),
            )        
        graphic=self.graphicsprovider.queryGraphic(oProps)
        image = self.document.createInstance("com.sun.star.text.GraphicObject")
        image.Graphic=graphic
        if name is not None:
            image.setName(name)
        image.AnchorType=anchortype
        #Calculate sizes
        if graphic.getType()==1:#Pixel
            ##Size100thMM sometimes is 0 in some images, but no Size
            if graphic.Size100thMM.Height==0 or graphic.Size100thMM.Height==0:
                debug(_("Problems to detect size in 100thmm, I'm trying to calculate it"))
            height100thmm=graphic.Size100thMM.Height if graphic.Size100thMM.Height!=0 else self.pixelsto100thmm(graphic.Size.Height)
            width100thmm=graphic.Size100thMM.Width if graphic.Size100thMM.Width!=0 else self.pixelsto100thmm(graphic.Size.Width)
            #print("100thmm problem ",  graphic.Size100thMM.Width,  graphic.Size100thMM.Height, width100thmm, height100thmm)
            
            
            
            if width is None and height is None:
                width=width100thmm/1000
                height=height100thmm/1000
            elif width is None and height is not None:
                width=round(height*width100thmm/height100thmm, 3)
            elif height is None and width is not None:
                height=round(width*height100thmm/width100thmm, 3)
            image.Size=Size(width*1000, height*1000)
        elif graphic.getType()==2:#Vector
            print("This is a vector graphic. TODO")
        elif graphic.getType()==0:#Empty
            print("This is a empty graphic. TODO")
        #print("Final size", width, height)
        return image
        
    ## @param filename_or_bytessequence_list List of filename path or byte_sequence
    ## @param width float in cm
    ## @param height float in cm
    def addImageParagraph(self, filename_or_bytessequence_list, width=None,  height=None, name=None, style="Illustration", linked=False):
        self.cursor.setPropertyValue("ParaStyleName", style)
        for filename_or_bytessequence in filename_or_bytessequence_list:
            self.document.Text.insertTextContent(self.cursor,self.textcontentImage(filename_or_bytessequence, width, height, "AS_CHARACTER", linked=linked), False)
        self.document.Text.insertControlCharacter(self.cursor, ControlCharacter.PARAGRAPH_BREAK, False)


    ## Table data, must be all strings (Not a spreadsheet)
    ## @params magings TRBL
    ## @params columnssize_percentage Por ejemplo para table 30,70, solo debo poner [30,70]
    ##   para table [10,10,80], solo debo poner [10,10]
    ## if none , rangos iguales
    def addTableParagraph(self, 
        data,  
        columnssize_percentages=None, 
        margins_top=0.3,
        margins_bottom=0.3,
        size=10,  
        width_percentage=100, 
        alignment="center",  
        paragraph_style="Standard", 
        style=None, 
        name=None
    ):        
        #Tabla definition
        table=self.document.createInstance("com.sun.star.text.TextTable")
        num_rows=len(data)
        if num_rows==0:
            return
        num_columns=len(data[0])
        table.initialize(num_rows, num_columns)
        if name is not None:
            table.TableName=name

        #Párrafo de la table
        self.cursor.setPropertyValue("ParaStyleName", paragraph_style)
        self.document.Text.insertTextContent(self.cursor,table,False)
        
        #Table data, must be all strings (Not a spreadsheet)
        if style is not None:
            table.TableTemplateName=style #MUST BE BEFORE TO OVERRIDE SOME PROPERTIES
        for row, row_data in enumerate(data):
            for column, cell_data in enumerate(row_data):
                cell=table.getCellByPosition( column, row)
                cursor = cell.createTextCursor()
                cursor.CharHeight=size
                if str(cell_data).startswith("-"):
                    cursor.setPropertyValue( "CharColor", 0xDD0000 )
                if style in ["Table0", "Table1", "Table1Total"] and row==0:
                    pass # Centered text for styles headers
                else:
                    if cell_data.__class__.__name__ in ["str"]:
                        cursor.ParaAdjust=LEFT
                    else:
                        cursor.ParaAdjust=RIGHT
                        cursor.ParaRightMargin=100
                cell.insertString(cursor, str(cell_data), False)
        
        #TAble width and style
        table.HoriOrient=2 #Centered
        table.TopMargin=margins_top*1000
        table.BottomMargin=margins_bottom*1000
        table.RelativeWidth=width_percentage #PARECE QUE ES SOLO DESCRIPTIVO
        #print("RelativeSum", table.TableColumnRelativeSum)
        #Columns width
        if columnssize_percentages is not None:
            separators=[]
            #print("Before", *self.printSeparators(table, table.TableColumnSeparators))
            for i, sep in enumerate(columnssize_percentages[:-1]):
                separator= createUnoStruct("com.sun.star.text.TableColumnSeparator")
                separator.Position=sum(columnssize_percentages[0:i+1])*100
                separator.IsVisible=True
                separators.append(separator)
            #print("Setting",  *self.printSeparators(table, separators))
            table.TableColumnSeparators=separators
            #print("After", *self.printSeparators(table, table.TableColumnSeparators))
            #print()

    def printSeparators(self, table, separators):
            r=[]
            for i in range(len(separators)):
                r.append(separators[i].Position)
            return (table.TableColumnRelativeSum, str(r))
            
    def paragraphBreak(self):
        self.document.Text.insertControlCharacter(self.cursor, ControlCharacter.PARAGRAPH_BREAK, False)
        
    ## Default styles are Landscape
    def pageBreak(self, style="Standard", page_before=True):
        if page_before is True:
            self.cursor.BreakType = PAGE_BEFORE 
        else:
            self.cursor.BreakType = PAGE_AFTER
            
        self.cursor.setPropertyValue("PageDescName", style)
        self.document.Text.insertString(self.cursor, "", False)



class ODS(ODF):
    def __init__(self, template=None, server=None):
        ODF.__init__(self, template, server)
        self._remove_default_sheet=True

    def getRemoveDefaultSheet(self):
        return self._remove_default_sheet
        
    ## @param b Boolean Sets if thiss class remove default sheet if it's empty
    def setRemoveDefaultSheet(self, b):
        self._remove_default_sheet=b
        
    ## Sets sheet pagestyle. If you use ODS_Standard you can set Default (Landscape) and Portrait
    ## Usefull when exporting to pdf
    def setSheetStyle(self, style):
        self.sheet.PageStyle=style
        
    def getSheetNames(self):        
        return self.document.Sheets.ElementNames

    ## Creates a new sheet at the end of the sheets
    ## @param if index is None it creates sheet at the end of the existing sheets
    def createSheet(self, name, index=None):
        for sheet in self.getSheetNames():
            if sheet.upper()==name.upper():
                raise exceptions.UnogeneratorException(_("ERROR: You can't create '{0}' sheet, because it already exists.").format(name)) 
        
        sheets=self.document.getSheets()
        if index is None:
            index=len(sheets)
        sheets.insertNewByName(name, index)
        self.setActiveSheet(index)
        
    def removeSheet(self, index):
        current=self.sheet.Name
        self.setActiveSheet(index)
        self.document.getSheets().removeByName(self.sheet.Name)
        self.document.getSheets().getByName(current)

    def setActiveSheet(self,  index):
        self.sheet_index=index
        self.sheet=self.document.getSheets().getByIndex(index)
        debug(f"Sheet '{self.sheet.Name}' ({self.sheet_index}) is now active")
        return self.sheet
    
    ## l measures are in cm can be float
    def setColumnsWidth(self, l):
        columns=self.sheet.getColumns()
        for i, width in enumerate(l):
            column=columns.getByIndex(i)
            width=width*1000
            column.Width=width ## Are in 1/100th of mm
            
    def setComment(self, coord, comment):
        coord=Coord.assertCoord(coord)
        celladdress= createUnoStruct("com.sun.star.table.CellAddress")
        celladdress.Sheet=self.sheet_index
        celladdress.Column=coord.letterIndex()
        celladdress.Row=coord.numberIndex()
        self.sheet.Annotations.insertNew(celladdress, comment)
            
    ## If addCell is used, preserves styles properties
    ## If you want to change these properties you must add parameters. Not developed
    def addCell(self, coord, o, color=None, outlined=None, alignment=None, decimals=None, bold=None):
        coord=Coord.assertCoord(coord)
        cell=self.sheet.getCellByPosition(coord.letterIndex(), coord.numberIndex())
        self.__object_to_cell(cell, o)

    def addRow(self, coord_start, list_o, formulas=True):        
        """
            Parameters:
                - formulas Boolean. If true formulas will be written as formula. If false as string
                - colors Color: Use color for all array, List of colors one for each cell
                - styles If None uses guest style. Else an array of styles
            Return: Range
        """
        coord_start=Coord.assertCoord(coord_start)
        
        if len(list_o)==0:
            debug(_("addRow is empty. Nothing to write. Ignoring..."))
            return None

        #Convert list_rows to valid dataarray
        r=[]
        for o in list_o:
            r.append(self.__object_to_dataarray_element(o))
        
        #Writes data fast
        range_indexes=[coord_start.letterIndex(), coord_start.numberIndex(), coord_start.letterIndex()+len(list_o)-1, coord_start.numberIndex()]
        range_uno=self.sheet.getCellRangeByPosition(*range_indexes)
        if formulas is True:
            self.__setFormulaArray(range_uno, [r, ])
        else:
            self.__setDataArray(range_uno, [r, ])
        return R.from_uno_range(range_uno)


    def addRowWithStyle(self, coord_start, list_o, colors=ColorsNamed.White,styles=None, formulas=True):        
        """
            Parameters:
                - formulas Boolean. If true formulas will be written as formula. If false as string
                - colors Color: Use color for all array, List of colors one for each cell
                - styles If None uses guest style. Else an array of styles
            Return: Range
        """
        coord_start=Coord.assertCoord(coord_start)        
        range_=self.addRow(coord_start, list_o, formulas)
        if range_ is None:
            return None
        range_uno=range_.uno_range(self.sheet)
        #Fast color:
        if styles is None:
            styles=[]
            for o in list_o:
                styles.append(guess_object_style(o))
            
        
        if colors.__class__==list:
            for i in range(len(list_o)):
                cell=self.sheet.getCellByPosition(coord_start.letterIndex()+i, coord_start.numberIndex())
                cell.setPropertyValue("CellBackColor", colors[i])
        else:
            range_uno.setPropertyValue("CellBackColor", colors)
        #Fast style:
        if styles.__class__==list:
            for i in range(len(list_o)):
                cell=self.sheet.getCellByPosition(coord_start.letterIndex()+i, coord_start.numberIndex())
                cell.setPropertyValue("CellStyle", styles[i])
        else:
            range_uno.setPropertyValue("CellStyle", styles)
        return range_
            


    def addColumn(self, coord_start, list_o, formulas=True):
        """
            Parameters:
                - formulas Boolean. If true formulas will be written as formula. If false as string
                - colors If None uses Wh
                - styles If None uses guest style. Else an array of styles
            Return: Range
        """
        coord_start=Coord.assertCoord(coord_start)
        
        if len(list_o)==0:
            return None

        #Convert list_o to valid dataarray
        r=[]
        for o in list_o:
            r.append([self.__object_to_dataarray_element(o), ])
        
        #Writes data fast
        range_indexes=[coord_start.letterIndex(), coord_start.numberIndex(), coord_start.letterIndex(), coord_start.numberIndex()+len(list_o)-1]
        range_=self.sheet.getCellRangeByPosition(*range_indexes)
        if formulas is True:
            self.__setFormulaArray(range_, r)
        else:
            self.__setDataArray(range_, r)
        return R.from_coords_indexes(*range_indexes)
            
    def addColumnWithStyle(self, coord_start, list_o, colors=ColorsNamed.White,styles=None, formulas=True):
        """
            Parameters:
                - formulas Boolean. If true formulas will be written as formula. If false as string
                - colors Color: Use color for all array, List of colors one for each cell
                - styles If None uses guest style. Else an array of styles
            Return: Range
        """
        coord_start=Coord.assertCoord(coord_start)
        range_=self.addColumn(coord_start, list_o, formulas)
        
        if range_ is None:
            return None
            
        range_uno=range_.uno_range(self.sheet)
        # Guess styles if none
        if styles is None:
            styles=[]
            for o in list_o:
                styles.append(guess_object_style(o))
        #Fast color:
        if colors.__class__==list:
            for i in range(len(list_o)):
                cell=self.sheet.getCellByPosition(coord_start.letterIndex(), coord_start.numberIndex()+i)
                cell.setPropertyValue("CellBackColor", colors[i])
        else:
            range_uno.setPropertyValue("CellBackColor", colors)
        #Fast style:
        if styles.__class__==list:
            for i in range(len(list_o)):
                cell=self.sheet.getCellByPosition(coord_start.letterIndex(), coord_start.numberIndex()+i)
                cell.setPropertyValue("CellStyle", styles[i])
        else:
            range_uno.setPropertyValue("CellStyle", styles)
        return range_
        
        
    
            
    def __setDataArray(self,  unorange,  array_):
        def contains_special_start(iterable, special_chars=('+', '=')):
            if isinstance(iterable, str):
                # Check if the string starts with any of the special characters
                return any(iterable.startswith(char) for char in special_chars)
            try:
                # Try iterating over the iterable
                for item in iterable:
                    if contains_special_start(item, special_chars):
                        return True
            except TypeError:
                # Not an iterable
                pass
            return False
        
        if contains_special_start(array_):
            debug(_("You're trying to add formulas to cells using setDataArray and they will be treated as strings. If you want formula's value use formulas=True in ODS methods"))
            debug(_("  + Range: {0}").format(unorange.AbsoluteName))
        unorange.setDataArray(array_)


    def __setFormulaArray(self,  unorange,  array_):
        unorange.setFormulaArray(array_)


    def addListOfRows(self, coord_start, list_rows, formulas=True):
        """
            Function used to add a big amount of cells to paste quickly
            This method is used when we want to add data without styles, or because we are using a styled template

            Parameters:
                - formulas Boolean. If true formulas will be written as formula. If false as string

            Return: Range
        """
        coord_start=Coord.assertCoord(coord_start) 
        
        rows=len(list_rows)
        if rows==0:
            columns=0
        else:
            columns=len(list_rows[0])
            
            
        if rows==0 or columns==0:
            debug(_("addListOfRowsWithStyle has {0} rows and {1} columns. Nothing to write. Ignoring...").format(rows, columns))
            return 
            
        #Convert list_rows to valid dataarray
        r=[]
        for row in list_rows:
            r_row=[]
            for o in row:
                r_row.append(self.__object_to_dataarray_element(o))
            r.append(r_row)

        #Writes data fast
        range_indexes=[coord_start.letterIndex(), coord_start.numberIndex(), coord_start.letterIndex()+columns-1, coord_start.numberIndex()+rows-1]
        range=self.sheet.getCellRangeByPosition(coord_start.letterIndex(), coord_start.numberIndex(), coord_start.letterIndex()+columns-1, coord_start.numberIndex()+rows-1)
        if formulas is True:
            self.__setFormulaArray(range, r)
        else:
            self.__setDataArray(range, r)
        return R.from_coords_indexes(*range_indexes)

    ## Function used to add a big amount of cells to paste quickly
    ## @param colors. List of column colors, one color, or None to use white
    ## @param styles. List of styles (columns) or None to guess them from first row
    ## @return range of the list_of_rows
    def addListOfRowsWithStyle(self, coord_start, list_rows, colors=ColorsNamed.White, styles=None,  formulas=True):
        """
            Function used to add a big amount of cells to paste quickly
            
            Parameters:
                - formulas Boolean. If true formulas will be written as formula. If false as string
                - colors. List of column colors, one color, or None to use white
                - styles. List of styles (columns) or None to guess them from first row
                
            Return: Range
        """
        coord_start=Coord.assertCoord(coord_start) 
        
        range_=self.addListOfRows(coord_start, list_rows, formulas)
        if range_ is None:
            return
        
        columns=range_.numColumns()
        rows=range_.numRows()

        # Parse colors.
        if colors.__class__.__name__=="list":
            colors=colors
        elif colors is None:
            colors=[ColorsNamed.White]*columns
        else: #one ColorsNamed
            colors=[colors]*columns
        
        # Parse styles
        if styles is None and rows>0:
            styles=[]
            for o in list_rows[0]:
                styles.append(guess_object_style(o))
        elif styles.__class__.__name__=="list":
            styles=styles
        else:
            styles=[styles]*columns
                
                
        if len(colors)!=columns:
            raise exceptions.UnogeneratorException(_("Colors must have the same number of items as data columns"))
        if len(styles)!=columns:
            raise exceptions.UnogeneratorException(_("Styles must have the same number of items as data columns"))
            
        #Create styles by columns cellranges
        if rows>0:
            for c, o in enumerate(list_rows[0]):
                columnrange=self.sheet.getCellRangeByPosition(coord_start.letterIndex()+c, coord_start.numberIndex(), coord_start.letterIndex()+c, coord_start.numberIndex()+rows-1)
                columnrange.setPropertyValue("CellStyle", styles[c])
                columnrange.setPropertyValue("CellBackColor", colors[c])                    
        return range_


    def addListOfColumns(self, coord_start, list_columns, formulas=True):
        """
            Parameters:
                - formulas Boolean. If true formulas will be written as formula. If false as string
                
            Return: Range
        """
        coord_start=Coord.assertCoord(coord_start) 
        list_rows=lol.lol_transposed(list_columns)
        return self.addListOfRows(coord_start, list_rows, formulas)


    ## @param style If None tries to guess it
    def addListOfColumnsWithStyle(self, coord_start, list_columns, colors=ColorsNamed.White, styles=None,  formulas=True):
        """
            Colors and styles are the colors of the first column. Code is different
            
            Parameters:
                - formulas Boolean. If true formulas will be written as formula. If false as string
            
            Return: Range
        """
        coord_start=Coord.assertCoord(coord_start) 
        
        range_=self.addListOfColumns(coord_start, list_columns, formulas)
        if range_ is None:
            return
        
        columns=range_.numColumns()
        rows=range_.numRows()

        # Parse colors.
        if colors.__class__.__name__=="list":
            colors=colors
        elif colors is None:
            colors=[ColorsNamed.White]*rows
        else: #one ColorsNamed
            colors=[colors]*rows
        
        # Parse styles
        if styles is None and rows>0:
            styles=[]
            for o in list_columns[0]:
                styles.append(guess_object_style(o))
        elif styles.__class__.__name__=="list":
            styles=styles
        else:
            styles=[styles]*rows
                
                
        if len(colors)!=rows:
            raise exceptions.UnogeneratorException(_("Colors must have the same number of items as data columns"))
        if len(styles)!=rows:
            raise exceptions.UnogeneratorException(_("Styles must have the same number of items as data columns"))
            
        #Create styles by columns cellranges
        if rows>0:
            for c, o in enumerate(list_columns[0]):
                columnrange=self.sheet.getCellRangeByPosition(coord_start.letterIndex(), coord_start.numberIndex()+c, coord_start.letterIndex()+columns-1, coord_start.numberIndex()+c)
                columnrange.setPropertyValue("CellStyle", styles[c])
                columnrange.setPropertyValue("CellBackColor", colors[c])                    
        return range_
            
    ## @param style If None tries to guess it
    ## @param rewritewrite If color is ColorsNamed.White, rewrites the color to White instead of ignoring it. Ignore it gains 0.200 ms
    ## THIS IS THE WAY TO CREATE FORMULAS
    def addCellWithStyle(self, coord, o, color=ColorsNamed.White, style=None):
        coord=Coord.assertCoord(coord)
        
        if style is None:
            style=guess_object_style(o)
                
        cell=self.sheet.getCellByPosition(coord.letterIndex(), coord.numberIndex())
        self.__object_to_cell(cell, o)
        cell.setPropertyValue("CellStyle", style)
        cell.setPropertyValue("CellBackColor", color)
        
    def setCellName(self, reference, name):
        """
            All of this names are document names
            Parameters:
                - reference -> str: Examples:   $Sheet2.$A$1 
        """
        cell=self.sheet.getCellRangeByName(reference).getCellAddress()
        self.document.NamedRanges.addNewByName(name, reference, cell, 0)
        
    def __object_to_cell(self, cell, o):
        if o.__class__.__name__ in ("int", ):
            cell.setValue(int(o))
        elif o.__class__.__name__ in ("str", ):
            if is_formula(o):
                cell.setFormula(o)
            else:
                cell.setString(o)
        elif o.__class__.__name__  == "datetime":
            cell.setValue(datetime2localc1989(o))
        elif o.__class__.__name__  == "date":
            cell.setValue(date2localc1989(o))
        elif o.__class__.__name__  == "time":
            cell.setValue(time2localc1989(o))
        elif o.__class__.__name__ in ("Percentage", ):
            if o.value is None:
                cell.setString("")
            else:
                cell.setValue(float(o.value))
        elif o.__class__.__name__ in ("Currency", "Money"):
            cell.setValue(float(o.amount))
        elif o.__class__.__name__ in ("Decimal",  "float"):
            cell.setValue(float(o))
        elif o.__class__.__name__ in ("bool", ):
            cell.setValue(int(o))
        elif o.__class__.__name__ in ("timedelta", ):
            cell.setValue(o.total_seconds())
        elif o is None:
            cell.setString("")
        else:
            cell.setString(str(o))
            print("MISSING", o.__class__.__name__)
            
    ## Function used to massive data_array creation
    def __object_to_dataarray_element(self, o):
        if o.__class__.__name__ in ("int", "float"):
            return o
        elif o.__class__.__name__ in ("timedelta", ):
            return o.total_seconds()
        elif o.__class__.__name__ in ("str", ):
            if is_formula(o):
                return o
            else:
                return o
        elif o.__class__.__name__  == "datetime":
            return datetime2localc1989(o)
        elif o.__class__.__name__  == "date":
            return date2localc1989(o)
        elif o.__class__.__name__  == "time":
            return time2localc1989(o)
        elif o.__class__.__name__ in ("Percentage", ):
            if o.value is None:
                 return ""
            else:
                return float(o.value)
        elif o.__class__.__name__ in ("Currency", "Money"):
            return float(o.amount)
        elif o.__class__.__name__ in ("Decimal", ):
            return float(o)
        elif o.__class__.__name__ in ("bool", ):
            return int(o)
        elif o is None:
            return ""
        else:
            return str(o)
            print("MISSING", o.__class__.__name__)

            
    def sortRange(self, range,  sortindex, ascending=True, casesensitive=True):
        range=R.assertRange(range)
        unorange=self.sheet.getCellRangeByName(range.string())
        sortfield=createUnoStruct("com.sun.star.table.TableSortField")
        sortfield.Field=sortindex
        sortfield.IsAscending=ascending
        sortfield.IsCaseSensitive=casesensitive
        sortDescr = unorange.createSortDescriptor()
        

        for prop in sortDescr:
            if prop.Name == 'SortFields':
                prop.Value = Any('[]com.sun.star.table.TableSortField', (sortfield,))
        # sort ...
        unorange.sort(sortDescr)

    def addCellMerged(self, range, o):
        range=R.assertRange(range)
        cell=self.sheet.getCellByPosition(range.c_start.letterIndex(), range.c_start.numberIndex())
        cellrange=self.sheet.getCellRangeByName(range.string())
        cellrange.merge(True)
        self.__object_to_cell(cell, o)
        return cell

    def addCellMergedWithStyle(self, range, o, color=ColorsNamed.White, style=None):
        cell=self.addCellMerged(range, o)
        if style is None:
            style=guess_object_style(o)
        cell.setPropertyValue("CellStyle", style)
        cell.setPropertyValue("CellBackColor", color)

    def freezeAndSelect(self, freeze, selected=None, topleft=None):
        freeze=Coord.assertCoord(freeze) 
        num_columns, num_rows=self.getSheetSize()
        self.document.getCurrentController().setActiveSheet(self.sheet)
        self.document.getCurrentController().freezeAtPosition(freeze.letterIndex(), freeze.numberIndex())

        if selected is None:
            selected=Coord.from_index(num_columns-1, num_rows-1)
        else:
            selected=Coord.assertCoord(selected)
        selectedcell=self.sheet.getCellByPosition(selected.letterIndex(), selected.numberIndex())
        self.document.getCurrentController().select(selectedcell)
        
        if topleft is None:
            #Get the select coord - 20 rows and - 10 columns
            minus_coord=selected.addRowCopy(-20).addColumnCopy(-10)
            if minus_coord.letterIndex()<=freeze.letterIndex():#letter
                letter=freeze.letter
            else:
                letter=minus_coord.letter
            if minus_coord.numberIndex()<=freeze.numberIndex():#number
                number=freeze.number
            else:
                number=minus_coord.number
            topleft=Coord.from_letters(letter, number)
        else:
            topleft=Coord.assertCoord(topleft)
        self.document.getCurrentController().setFirstVisibleColumn(topleft.letterIndex())
        self.document.getCurrentController().setFirstVisibleRow(topleft.numberIndex())
        
        
    def getValue(self, coord,  detailed=False):
        """
            Gets the value from a coord
            If detailed is True returns a dict with detailed information
        """
        coord=Coord.assertCoord(coord)
        return self.getValueByPosition(coord.letterIndex(), coord.numberIndex(), detailed)
        
    def getValueByPosition(self, letter_index, number_index,  detailed=False):
        """
            Gets the value from a position A1= (0,0)
            If detailed is True returns a dict with detailed information
        """
        r=self.__cell_to_object(self.sheet.getCellByPosition(letter_index, number_index), detailed)
        return r

    ## Returns a list of rows with the values of the sheet
    ## @param sheet_index Integer index of the sheet
    ## @param skip_up int. Number of rows to skip at the begining of the list of rows (lor)
    ## @param skip_down int. Number of rows to skip at the end of the list of rows (lor)
    ## @param detailed Returns a dict {'value': 'A1', 'string': 'A1', 'style': 'Normal', 'class': 'str', 'is_formula': False, 'formula': None}
    ## @param casts List of strings with style names, to cast string and float to objects. Allowed values
    ##      "int", "str", "Decimal", "float", "Percentage", "USD", "EUR",
    ##      The cast columns, so you need to have the same cast items as columns in range
    ## If detailed is True cast is ignored
    ## @return Returns a list of rows of object values
    def getValues(self, skip_up=0, skip_down=0, skip_left=0, skip_right=0, cast=None, detailed=False):
        range_=self.getSheetRange()
        if skip_up>0:
           range_=range_.addRowBefore(-skip_up)
        if skip_down>0:
           range_=range_.addRowAfter(-skip_down)
        if skip_left>0:
           range_=range_.addColumnBefore(-skip_left)
        if skip_right>0:
           range_=range_.addColumnAfter(-skip_right)
        return self.getValuesByRange(range_, cast,  detailed)

    ## Get values by range. If detailed is False, it uses getDataArray method: Dates and datetimes are float. You can convert them using commons.localc1989 methods
    ## @param sheet_index Integer index of the sheet
    ## @param range_ Range object to get values. If None returns all values from sheet
    ## @param detailed Returns a dict {'value': 'A1', 'string': 'A1', 'style': 'Normal', 'class': 'str', 'is_formula': False, 'formula': None}
    ## @param casts List of strings with style names, to cast string and float to objects. Allowed values
    ##      "int", "str", "Decimal", "float", "Percentage", "USD", "EUR",
    ##      The cast columns, so you need to have the same cast items as columns in range
    ## If detailed is True cast is ignored
    ## @return Returns a list of rows of object values
    def getValuesByRange(self, range_,  cast=None, detailed=False):
        range_=R.assertRange(range_)
        
        if detailed is True:#position by position
            r=[]
            for row_indexes in range_.indexes_list():
                rrow=[]
                for column_index,  row_index in row_indexes:
                    rrow.append(self.getValueByPosition(column_index, row_index, detailed))
                r.append(rrow)
            return r
        else:
            range_=R.assertRange(range_)
            range_uno=self.sheet.getCellRangeByPosition(range_.c_start.letterIndex(), range_.c_start.numberIndex(), range_.c_end.letterIndex(), range_.c_end.numberIndex())
            tupleoftuples=range_uno.getDataArray()
            if cast is not None:
                #Reads data fast
                lor=[]
                for tuple_ in tupleoftuples:
                    lor_row=[]
                    for i, tuple_value in enumerate(tuple_):
                        lor_row.append(string_float2object(tuple_value, cast[i]))
                    lor.append(lor_row)
                return lor
            else: #cast false
                return tupleoftuples
       
    ## @param sheet_index Integer index of the sheet
    ## @param column_letter Letter of the column to get values
    ## @param skip Integer Number of top rows to skip in the result
    ## @param detailed Returns a dict {'value': 'A1', 'string': 'A1', 'style': 'Normal', 'class': 'str', 'is_formula': False, 'formula': None}
    ## @param casts List of strings with style names, to cast string and float to objects. Allowed values
    ##      "int", "str", "Decimal", "float", "Percentage", "USD", "EUR",
    ##      The cast columns, so you need to have the same cast items as columns in range
    ## If detailed is True cast is ignored
    ## @return List of values
    def getValuesByColumn(self, column_letter, skip_up=0, skip_down=0,  cast=None, detailed=False):
        range_=self.getSheetRange()
        range_.c_start.letter=column_letter
        range_.c_end.letter=column_letter
        if skip_up>0:
           range_=range_.addRowBefore(-skip_up)
        if skip_down>0:
           range_=range_.addRowAfter(-skip_down)
        lor=self.getValuesByRange(range_, cast, detailed)
        #Transform to list
        r=[]
        for o in lor:
            r.append(o[0])
        return r

    ## @param sheet_index Integer index of the sheet
    ## @param row_number String Number of the row to get values
    ## @param skip Integer Number of top rows to skip in the result
    ## @param detailed Returns a dict {'value': 'A1', 'string': 'A1', 'style': 'Normal', 'class': 'str', 'is_formula': False, 'formula': None}
    ## @param casts List of strings with style names, to cast string and float to objects. Allowed values
    ##      "int", "str", "Decimal", "float", "Percentage", "USD", "EUR",
    ##      The cast columns, so you need to have the same cast items as columns in range
    ## If detailed is True cast is ignored
    ## @return List of values
    def getValuesByRow(self, row_number, skip_left=0, skip_right=0,  cast=None,  detailed=False):
        range_=self.getSheetRange()
        if skip_left>0:
           range_=range_.addColumnBefore(-skip_left)
        if skip_right>0:
           range_=range_.addColumnAfter(-skip_right)
        lor=self.getValuesByRange(range_, cast, detailed)
        return lor[0]



    ## If formula  or without known style returns a tuple (value, formula)
    ## if detailed return a tuple (value, formula, style)
    def __cell_to_object(self, cell, detailed=False):
        isformula=is_formula(cell.getFormula())
        if isformula:
            value=cell.getValue()
        if cell.CellStyle=="Bool":
            value=bool(cell.getValue())
        elif cell.CellStyle in ["Datetime"]:
            value=casts.str2dtnaive(cell.getString(),"%Y-%m-%d %H:%M:%S.")
        elif cell.CellStyle in ["Date"]:
            value=casts.str2date(cell.getString())
        elif cell.CellStyle in ["Time"]:
            value=casts.str2time(cell.getString(), "HH:MM:SS")
        elif cell.CellStyle in ["Float2", "Float6"]:
            value=cell.getValue()
        elif cell.CellStyle in ["Integer"]:
            value=int(cell.getValue())
        elif cell.CellStyle in ["EUR", "USD"]:
            value=Currency(cell.getValue(), cell.CellStyle)
        elif cell.CellStyle in ["Percentage"]:
            value=Percentage(cell.getValue(), 1)
        else:
            value=cell.getString()
        
        if detailed is False:
            return value
        else:
            formula = cell.getFormula() if isformula else None
            return {"value":value, "string":  cell.getString(),  "style":cell.CellStyle, "class": value.__class__.__name__, "is_formula": isformula, "formula": formula}

    ## Return a Range object with the limits of the index sheet
    def getSheetRange(self):
        columns,  rows=self.getSheetSize()
        endcoord=Coord("A1").addRow(rows-1).addColumn(columns-1)
        return R("A1:" + endcoord.string())
        

    ## Returns (columnsNumber, rowsNumber
    def getSheetSize(self):
        data=self.sheet.getData()
        if len(data)==0:
            return 0, 0
        else:
            return len(data[0]), len(data)
            
    ## Method to remove default sheet if empty
    def removeDefaultSheet(self):
        sheets=self.document.getSheets()
        for sheet in sheets:
            if sheet.getName() in ("Hoja1", "Sheet1"):
                data=sheet.getData()
                d=sheet.getCellByPosition(0, 0)
                if len(data)==1 and len(data[0])==1 and d.Value==0 and len(sheets)>1: #Empty and A1 value 0 and number sheets >1
                    self.document.getSheets().removeByName(sheet.getName())
            
    def save(self, filename, overwrite_template=False):
        if self.getRemoveDefaultSheet() is True:
            self.removeDefaultSheet()
            
        if filename==self.template and overwrite_template is False:
            print(_("You can't use the same filename as your template or you will overwrite it."))
            print(_("You can force it, setting overwrite_template paramter to True"))
            print(_("Document hasn't been saved."))
            return
        if filename.endswith(".ods") is False:
            print(_("Filename extension must be 'ods'."))
            print(_("Document hasn't been saved."))
            return
        ## SAVE FILE
        args=(
            PropertyValue('FilterName', 0, 'calc8', 0),
            PropertyValue('Overwrite', 0, True, 0),
        )
        with TemporaryDirectory() as tmpdirname:
                tempfile=f"{tmpdirname}/{path.basename(filename)}"
                self.document.storeAsURL(systemPathToFileUrl(tempfile), args)
                makedirs(path.dirname(path.abspath(filename)), exist_ok=True)
                copyfile(tempfile, filename)

    ## If you want to set one page to portrait use setSheetStyle
    ## @filename
    ## @page_per_sheet If True exports pdf with a pdf page per sheet. If False makes a standard export
    def export_pdf(self, filename, page_per_sheet=True):
        if self.getRemoveDefaultSheet() is True:
            self.removeDefaultSheet()
        
        if filename.endswith(".pdf") is False:
            print(_("Filename extension must be 'pdf'."))
            print(_("Document hasn't been saved."))
            return
            
        if page_per_sheet is True:
            oStyleFamilies =self.document.getStyleFamilies()
            oObj_1 = oStyleFamilies.getByName("PageStyles")
            oObj_2 = oObj_1.getByName("Default")
            oObj_2.ScaleToPagesX= 1
            oObj_2.ScaleToPagesY = 1
            args=(
                PropertyValue('FilterName',0,'calc_pdf_Export',0),
                PropertyValue('Overwrite',0,True,0),
            )
        else: #page_per_sheet False
            args=(
                PropertyValue('FilterName',0,'calc_pdf_Export',0),
                PropertyValue('Overwrite',0,True,0),
            )
        
        with TemporaryDirectory() as tmpdirname:
            tempfile=f"{tmpdirname}/{path.basename(filename)}"
            self.document.storeToURL(systemPathToFileUrl(tempfile), args)
            makedirs(path.dirname(path.abspath(filename)), exist_ok=True)
            copyfile(tempfile, filename)


    def export_xlsx(self, filename):
        if self.getRemoveDefaultSheet() is True:
            self.removeDefaultSheet()

        if filename.endswith(".xlsx") is False:
            print(_("Filename extension must be 'xlsx'."))
            print(_("Document hasn't been saved."))
            return
        args=(
            PropertyValue('FilterName',0,'Calc MS Excel 2007 XML',0),
            PropertyValue('Overwrite',0,True,0),
        )
        with TemporaryDirectory() as tmpdirname:
            tempfile=f"{tmpdirname}/{path.basename(filename)}"
            self.document.storeToURL(systemPathToFileUrl(tempfile), args)
            makedirs(path.dirname(path.abspath(filename)), exist_ok=True)
            copyfile(tempfile, filename)
            
    def setColorScale(self, range):
        range=R.assertRange(range)
        myCells=self.sheet.getCellRangeByName(range.string())
#        print(unorange, dir(unorange))
#        print(self.sheet, dir(self.sheet))
        myConditionalFormat = myCells.ConditionalFormat
        args=(
            PropertyValue('Operator',0, COLORSCALE,0),
        )
        myConditionalFormat.clear()
        myConditionalFormat.addNew(args)
        myCells.ConditionalFormat = myConditionalFormat
#        self.ws_current.conditional_formatting.add(range, 
#                            openpyxl.formatting.rule.ColorScaleRule(
#                                                start_type='percentile', start_value=0, start_color='00FF00',
#                                                mid_type='percentile', mid_value=50, mid_color='FFFFFF',
#                                                end_type='percentile', end_value=100, end_color='FF0000'
#                                                )
#                                            )

    def toDictionaryOfDetailedValues(self):
        """
            Converts document to a dictionary with detailed values of every sheets
            Has two keys:
                - sheets. A list of dictionaries
                - dictionary. A dictionary with keys (sheet_name, coord_string)
            
            For example: {
                'dictionary': {
                    ('One', 'A1'): {'value': 2, 'string': '2', 'style': 'Integer', 'class': 'int', 'is_formula': False, 'formula': None}, 
                    ('One', 'A2'): {'value': '', 'string': '', 'style': 'Default', 'class': 'str', 'is_formula': False, 'formula': None}, 
                    ('One', 'A3'): {'value': 4, 'string': '4', 'style': 'Integer', 'class': 'int', 'is_formula': True, 'formula': '=2*hola'}, 
                    ('Two', 'A1'): {'value': 5, 'string': '5', 'style': 'Integer', 'class': 'int', 'is_formula': False, 'formula': None}
                }, 
                'sheets': [
                    {'name': 'One', 'columns': 1, 'rows': 3}, 
                    {'name': 'Two', 'columns': 1, 'rows': 1}
                ]
            }

            This method retains more exact values (types) with ODS_Standard although woks fine with ODS
        """
        r={"dictionary":{}, "sheets":[]}
        for i, sheet_name in enumerate(self.getSheetNames()):
            self.setActiveSheet(i)
            sheet_values=self.getValues(detailed=True)
            for number_index, row in enumerate(sheet_values):
                for letter_index, cell in enumerate(row):
                    r["dictionary"][(sheet_name, Coord.from_index(letter_index, number_index).string())]=cell
            r["sheets"].append({"name":sheet_name, "columns": letter_index+1,  "rows": number_index+1})
        return r

class ODS_Standard(ODS):
    def __init__(self, server=None):
        ODS.__init__(self, files('unogenerator') / 'templates/standard.ods', server)

class ODT_Standard(ODT):
    def __init__(self, server=None):
        ODT.__init__(self, files('unogenerator') / 'templates/standard.odt', server)
        self.deleteAll()





## @namespace unogenerator.demo
## @brief Generate ODF example files
from uno import getComponentContext
getComponentContext()
import argparse
from collections import OrderedDict
from concurrent.futures import ProcessPoolExecutor, as_completed,  ThreadPoolExecutor
from datetime import datetime, date, timedelta
from gettext import translation
from logging import info
from importlib.resources import files
from os import system
from pydicts.currency import Currency
from pydicts.percentage import Percentage
from unogenerator import ODT_Standard, ODS_Standard, __version__,  commons, ColorsNamed, Coord, LibreofficeServer
from unogenerator.helpers import helper_title_values_total_row,helper_title_values_total_column, helper_totals_row, helper_totals_from_range, helper_list_of_ordereddicts, helper_list_of_dicts, helper_list_of_ordereddicts_with_totals, helper_ods_sheet_stylenames, helper_split_big_listofrows
from tqdm import tqdm

try:
    t=translation('unogenerator', files("unogenerator") / 'locale')
    _=t.gettext
except:
    _=str

type_choices=[ "SEQUENTIAL",  "CONCURRENT_PROCESS",  "CONCURRENT_THREADS", "COMMONSERVER_SEQUENTIAL","COMMONSERVER_CONCURRENT_PROCESS","COMMONSERVER_CONCURRENT_THREADS" ]

## If arguments is None, launches with sys.argc parameters. Entry point is toomanyfiles:main
## You can call with main(['--pretend']). It's equivalento to os.system('program --pretend')
## @param arguments is an array with parser arguments. For example: ['--argument','9']. 
def demo(arguments=None):
    parser=argparse.ArgumentParser(prog='unogenerator', description=_('Create example files using unogenerator module'), epilog=commons.argparse_epilog(), formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--debug', help=_("Debug program information"), choices=["DEBUG","INFO","WARNING","ERROR","CRITICAL"], default="ERROR")
    group= parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--create', help="Create demo files", action="store_true",default=False)
    group.add_argument('--remove', help="Remove demo files", action="store_true", default=False)
    group.add_argument('--benchmark', help="Executes all types to compare its benchmark", action="store_true", default=False)
    parser.add_argument('--type', help="Debug program information", choices=type_choices,  default="COMMONSERVER_CONCURRENT_PROCESS")
    args=parser.parse_args(arguments)
    commons.addDebugSystem(args.debug)
    demo_command(args.create, args.remove, args.benchmark, args.type)

def demo_command(create, remove, benchmark, type):
    languages=['es', 'en',  'ro',  'fr']
        
    if benchmark is True:
        for type in type_choices:
            #demo_command(True, False,  False,  type)
            system(f"unogenerator_demo --create --type {type}")

    if remove==True:
            for language in languages:
                commons.remove_without_errors(f"unogenerator_documentation_{language}.odt")
                commons.remove_without_errors(f"unogenerator_documentation_{language}.docx")
                commons.remove_without_errors(f"unogenerator_documentation_{language}.pdf")
                commons.remove_without_errors(f"unogenerator_example_{language}.ods")
                commons.remove_without_errors(f"unogenerator_example_{language}.xlsx")
                commons.remove_without_errors(f"unogenerator_example_{language}.pdf")

    if create==True:
        start=datetime.now()
        instances=3
        total_documents=len(languages)*2
        
        if type=="CONCURRENT_PROCESS":            
            futures=[]
            print(_("Launching demo with {0} workers without common server using concurrent processes").format(instances))

            with ProcessPoolExecutor(max_workers=instances) as executor:
                with tqdm(total=total_documents) as progress:
                    for language in languages:
                        future=executor.submit(demo_ods_standard, language, None)
                        future.add_done_callback(lambda p: progress.update())
                        futures.append(future)
                        future=executor.submit(demo_odt_standard, language,  None)
                        future.add_done_callback(lambda p: progress.update())
                        futures.append(future)

                    for future in as_completed(futures):
                        future.result()

            results = []
            for future in futures:
                result = future.result()
                results.append(result)    

        elif type=="COMMONSERVER_CONCURRENT_PROCESS":            
            futures=[]
            print(_("Launching demo with {0} workers with common server using concurrent processes").format(instances))

            with LibreofficeServer() as server: #FALLA POR PICCKING
                with ProcessPoolExecutor(max_workers=instances) as executor:
                    with tqdm(total=total_documents) as progress:
                            for language in languages:
                                future=executor.submit(demo_ods_standard, language, server)
                                future.add_done_callback(lambda p: progress.update())
                                futures.append(future)
                                future=executor.submit(demo_odt_standard, language,  server)
                                future.add_done_callback(lambda p: progress.update())
                                futures.append(future)

                            for future in as_completed(futures):
                                future.result()

                results = []
                for future in futures:
                    result = future.result()
                    results.append(result)

        elif type=="CONCURRENT_THREADS":            
            futures=[]
            print(_("Launching demo with {0} workers without common server using concurrent threads").format(instances))

            with ThreadPoolExecutor(max_workers=instances) as executor:
                with tqdm(total=total_documents) as progress:
                    for language in languages:
                        future=executor.submit(demo_ods_standard, language, None)
                        future.add_done_callback(lambda p: progress.update())
                        futures.append(future)
                        future=executor.submit(demo_odt_standard, language,  None)
                        future.add_done_callback(lambda p: progress.update())
                        futures.append(future)

                    for future in as_completed(futures):
                        future.result()

                results = []
                for future in futures:
                    result = future.result()
                    results.append(result)

        elif type=="COMMONSERVER_CONCURRENT_THREADS":            
            futures=[]
            print(_("Launching demo with {0} workers with common server using concurrent threads").format(instances))

            with LibreofficeServer() as server: #FALLA POR PICCKING
                with ThreadPoolExecutor(max_workers=instances) as executor:
                    with tqdm(total=total_documents) as progress:
                        for language in languages:
                            future=executor.submit(demo_ods_standard, language, server)
                            future.add_done_callback(lambda p: progress.update())
                            futures.append(future)
                            future=executor.submit(demo_odt_standard, language,  server)
                            future.add_done_callback(lambda p: progress.update())
                            futures.append(future)

                        for future in as_completed(futures):
                            future.result()

                results = []
                for future in futures:
                    result = future.result()
                    results.append(result)

        elif type=="COMMONSERVER_SEQUENTIAL":
            with LibreofficeServer() as server:
                print(_("Launching demo with one common server sequentially"))
                with tqdm(total=total_documents) as progress:
                    for language in languages:
                        demo_ods_standard(language, server)
                        progress.update()
                        demo_odt_standard(language,  server)       
                        progress.update()
                        
        elif type=="SEQUENTIAL":
                print(_("Launching demo without one common server sequentially"))
                with tqdm(total=total_documents) as progress:
                    for language in languages:
                        demo_ods_standard(language, None)
                        progress.update()
                        demo_odt_standard(language,  None)       
                        progress.update()     
            
        print(_("All process took {}".format(datetime.now()-start)))

       
def demo_ods_standard(language, server):
    lang1=translation('unogenerator', files("unogenerator") / 'locale', languages=[language])
    lang1.install()
    _=lang1.gettext
    
    with ODS_Standard(server=server) as doc:
        doc.setMetadata(
            _("UnoGenerator ODS example"),  
            _("Demo with ODS class"), 
            "Turulomio", 
            _("This file have been generated with UnoGenerator-{0}. You can see UnoGenerator main page in https://github.com/turulomio/unogenerator").format(__version__), 
            ["unogenerator", "demo", "files"]
        )
        doc.createSheet("Styles")
        doc.setColumnsWidth([3.5, 5, 2, 2, 2, 2, 2, 5, 5, 3, 3])
        
        doc.setSheetStyle("Portrait")
        doc.setCellName("A1",  "MYNAME")
        
        doc.addCellWithStyle("A1", _("Style name"), ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("B1", _("Date and time"), ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("C1", _("Date"), ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("D1", _("Integer"), ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("E1", _("Euros"), ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("F1", _("Dollars"), ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("G1", _("Percentage"), ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("H1", _("Number with 2 decimals"), ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("I1", _("Number with 6 decimals"), ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("J1", _("Time"), ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("K1", _("Boolean"), ColorsNamed.Orange, "BoldCenter")
        colors_list=([a for a in dir(ColorsNamed()) if not a.startswith('__')])
        for row, color_str in enumerate(colors_list):
            color_key=getattr(ColorsNamed(), color_str)
            doc.addCellWithStyle(Coord("A2").addRow(row), color_str, color_key, "Bold")
            doc.addCellWithStyle(Coord("B2").addRow(row), datetime.now(), color_key, "Datetime")
            doc.addCellWithStyle(Coord("C2").addRow(row), date.today(), color_key, "Date")
            doc.addCellWithStyle(Coord("D2").addRow(row), pow(-1, row)*-10000000, color_key, "Integer")
            doc.addCellWithStyle(Coord("E2").addRow(row), Currency(pow(-1, row)*12.56, "EUR"), color_key, "EUR")
            doc.addCellWithStyle(Coord("F2").addRow(row), Currency(pow(-1, row)*12345.56, "USD"), color_key, "USD")
            doc.addCellWithStyle(Coord("G2").addRow(row), Percentage(pow(-1, row)*1, 3), color_key,  "Percentage")
            doc.addCellWithStyle(Coord("H2").addRow(row), pow(-1, row)*123456789.121212, color_key, "Float6")
            doc.addCellWithStyle(Coord("I2").addRow(row), pow(-1, row)*-12.121212, color_key, "Float2")
            doc.addCellWithStyle(Coord("J2").addRow(row), (datetime.now()+timedelta(seconds=3600*12*row)).time(), color_key, "Time")
            doc.addCellWithStyle(Coord("K2").addRow(row), bool(row%2), color_key, "Bool")

        doc.addCellWithStyle(Coord("E2").addRow(row+1),f"=sum(E2:{Coord('E2').addRow(row).string()})", ColorsNamed.GrayLight, "EUR" )
        doc.addCellMergedWithStyle("E15:K15", "Merge proof", ColorsNamed.Yellow, style="BoldCenter")
        doc.setComment("B14", "This is nice comment")
        
        doc.freezeAndSelect("B2")
        
        ## List of rows
        doc.createSheet("List of rows or columns")
        
                
        doc.addCellMergedWithStyle("A1:C1","List of rows with helper_totals_row", ColorsNamed.Orange, "BoldCenter")
        range_=doc.addListOfRowsWithStyle("A2", [[1,2,3],[4,5,6],[7,8,9]], ColorsNamed.White)
        helper_totals_row(doc, range_.c_start.addRowCopy(range_.numRows()), ["#SUM"]*3, styles=None, row_from="2", row_to="4")
        
        doc.addCellMergedWithStyle("A8:C8","List of columns with helper_totals_row", ColorsNamed.Orange, "BoldCenter")
        range_=doc.addListOfColumnsWithStyle("A9", [[1,2,3],[4,5,6],[7,8,9]], ColorsNamed.White)
        helper_totals_row(doc, range_.c_start.addRowCopy(range_.numRows()), ["#SUM"]*3, styles=None, row_from="9", row_to="11")

        doc.addCellMergedWithStyle("A15:E15","List of rows with helper_totals_from_range in rows and columns", ColorsNamed.Orange, "BoldCenter")
        range_=doc.addListOfRowsWithStyle("A16", [["A",12000,2,3, 6],["B",1020,5,6, 7],["C",20404,8,9, 8]], ColorsNamed.White)
        helper_totals_from_range(doc, range_, totalcolumns=True, totalrows=True)
        
        
        doc.addCellMergedWithStyle("A22:E22","List of rows with helper_totals_from_range in rows", ColorsNamed.Orange, "BoldCenter")
        range_=doc.addListOfRowsWithStyle("A23", [["A",12000,2,3, 6],["B",1020,5,6, 7],["C",20404,8,9, 8]], ColorsNamed.White)
        helper_totals_from_range(doc, range_.addColumnBefore(-1), totalcolumns=False, totalrows=True) #Removes one column to filter first alphanumerical column

        doc.addCellMergedWithStyle("A29:E29","List of rows with helper_totals_from_range in columns", ColorsNamed.Orange, "BoldCenter")
        range_=doc.addListOfRowsWithStyle("A30", [["A",12000,2,3, 6],["B",1020,5,6, 7],["C",20404,8,9, 8]], ColorsNamed.White)
        helper_totals_from_range(doc, range_.addColumnBefore(-1), totalcolumns=True, totalrows=False)
        
        doc.addCellMergedWithStyle("A35:E35","List of rows with helper_totals_from_range in rows showing", ColorsNamed.Orange, "BoldCenter")
        range_=doc.addListOfRowsWithStyle("A36", [["A",12000,2,3, 6],["B",1020,5,6, 7],["C",20404,8,9, 8]], ColorsNamed.White)
        helper_totals_from_range(doc, range_.addColumnBefore(-1), totalcolumns=False, totalrows=True, showing=True)

        doc.addCellMergedWithStyle("A42:E42","List of rows with helper_totals_from_range in columns showing", ColorsNamed.Orange, "BoldCenter")
        range_=doc.addListOfRowsWithStyle("A43", [["A",12000,2,3, 6],["B",1020,5,6, 7],["C",20404,8,9, 8]], ColorsNamed.White)
        helper_totals_from_range(doc, range_.addColumnBefore(-1), totalcolumns=True, totalrows=False, showing=True)

        

        ## HELPERS
        doc.createSheet("Helpers")
        doc.setSheetStyle("Portrait")
        doc.addCellMergedWithStyle("A1:E1","Helper values with total (horizontal)", ColorsNamed.Orange, "BoldCenter")
        helper_title_values_total_row(doc, "A2", "Suma 3", [1,2,3])

        doc.addCellMergedWithStyle("A4:A9","Helper values with total (vertical)", ColorsNamed.Orange, "VerticalBoldCenter")
        helper_title_values_total_column(doc, "B4", "Suma 3", [1,2,3, 4])

        
        doc.addCellMergedWithStyle("A23:B23","List of ordered dictionaries", ColorsNamed.Orange, "BoldCenter")
        lod=[]
        lod.append(OrderedDict({"Singer": "Elvis",  "Song": "Fever" }))
        lod.append(OrderedDict({"Singer": "Roy Orbison",  "Song": "Blue angel" }))
        helper_list_of_ordereddicts(doc, "A24",  lod, columns_header=1)
        
        doc.addCellMergedWithStyle("A28:B28","List of dictionaries", ColorsNamed.Orange, "BoldCenter")
        helper_list_of_dicts(doc, "A29",  lod, keys=["Song",  "Singer"])
        
        doc.addCellMergedWithStyle("A33:D33","List of ordered dictionaries one method with totals", ColorsNamed.Orange, "BoldCenter")
        lod=[]
        lod.append(OrderedDict({"Singer": "Elvis",  "Songs": 10000 , "Albums": 100}))
        lod.append(OrderedDict({"Singer": "Roy Orbison",  "Songs": 100,  "Albums": 20 }))
        helper_list_of_ordereddicts_with_totals(doc, "A34",  lod, columns_header=1)
        
        ##Sort
        doc.createSheet("Sort")
        l=[7, 3, 2, 5, 6, 0, 9, 4, 10]
        doc.addCellWithStyle("A1",  "Unsorted", ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("B1", "Sorted ASC", ColorsNamed.Orange, "BoldCenter")
        doc.addCellWithStyle("C1", "Sorted DESC", ColorsNamed.Orange, "BoldCenter")
        doc.addColumnWithStyle("A2", l)
        doc.addColumnWithStyle("B2", l)
        doc.addColumnWithStyle("C2", l)
        doc.sortRange("B2:B10",  0)
        doc.sortRange("C2:C10",  0, False)
        
        ## Split big LOR
        lor=[]
        for i in range(1000):
            lor.append([i, _("String")+" "+ str(i), datetime.now()])
            
        helper_split_big_listofrows(doc, "Splits in 400 rows", lor, ["Integer", "String", "Datetime"], columns_width=[2, 5, 5],  max_rows=400)

        
        ## Sheet with all styles names
        helper_ods_sheet_stylenames(doc)

        doc.save(f"unogenerator_example_{language}.ods")
        doc.export_xlsx(f"unogenerator_example_{language}.xlsx")
        doc.export_pdf(f"unogenerator_example_{language}.pdf")
    
    r= _("unogenerator_example_{0}.ods took {1} in {2}").format(language, datetime.now()-doc.start, doc.server.port)
    info(r)
    return r
    
    
def demo_odt_standard(language, server):
    lang1=translation('unogenerator', files("unogenerator") / 'locale', languages=[language])
    lang1.install()
    _=lang1.gettext

    with ODT_Standard(server=server) as doc:
        doc.setMetadata(
            _("UnoGenerator documentation"),  
            _("UnoGenerator python module documentation"), 
            "Turulomio", 
            _("This file have been generated with UnoGenerator-{0}. You can see UnoGenerator main page in https://github.com/turulomio/unogenerator").format(__version__), 
            ["unogenerator", "demo", "files"]
        )
        
        
        doc.addParagraph(_("UnoGenerator documentation"), "Title")
        doc.addParagraph(_("Version: {0}").format(__version__), "Subtitle")

        doc.addImageParagraph([files('unogenerator') / 'images/unogenerator.png', ], 4, 4, "Illustration", linked=False)


        
        doc.addParagraph(_("Introduction"),  "Heading 1")
        
        doc.addParagraph(
            _("UnoGenerator uses Libreoffice UNO API python bindings to generate documents.") +" " +
            _("So in order to use, you need to launch a --headless LibreOffice instance.") + " "+
            _("UnoGenerator make this unattended for you in each ODF (ODS and ODT) instance.") + " " +
            _("However if you wish, you can do this programmatically using LibreofficeServer class to reuse it in serveral ODF instances to improve documents generation speed."),
            "Standard"
        )

        doc.addParagraph(
            _("UnoGenerator has standard templates to help you with edition, although you can use your own templates.") +" " + 
            _("You can edit this one or create your own.")  +" " +
            _("This document has been created with 'standard.odt' files that you can find inside this python module."), 
            "Standard"
        )
            
        doc.addParagraph(_("Installation"), "Heading 2")
        doc.addParagraph(_("You can use pip to install this python package:") ,  "Standard")
        doc.addParagraph("""pip install unogenerator"""    , "Code")
        doc.addParagraph(_("ODT 'Hello World' example"), "Heading 2")
        doc.addParagraph(_("This is a Hello World example. You get the example in odt, docx and pdf formats:") ,  "Standard")
        doc.addParagraph("""from unogenerator import ODT_Standard
with ODT_Standard() as doc:
    doc.addParagraph("Hello World", "Heading 1")
    doc.addParagraph("Easy, isn't it","Standard")
    doc.save("hello_world.odt")
    doc.export_docx("hello_world.docx")
    doc.export_pdf("hello_world.pdf")"""    , "Code")
        doc.pageBreak()
        
        doc.addParagraph(_("ODT"), "Heading 1")
        doc.addParagraph(
            _("ODT files can be quickly generated with UnoGenerator.") + " " + 
            _("There is a predefined template in code called 'standard.odt' to help you with edition."),  
            "Standard"
        )
        
        
        doc.addParagraph(_("Calling the ODT constructor"), "Heading 2")
        doc.addParagraph(_("You can call ODT constructor in this ways:") , "Standard")
            
        doc.addParagraph(
            _("ODT with standard template (Recomended).") + " " + 
            _("There is a predefined template in code called 'standard.odt', inside this python module, to help you with edition, although you can use your own ones.") +" "+
            _("With this mode you can create new documents"), 
            "BulletsLevel1"
        )
        
        doc.addParagraph("""from unogenerator import ODT_Standard
doc=ODT_Standard()"""    , "Code")

        doc.addParagraph(
            _("ODT with template or file (Recomended).") + " " + 
            _("With this mode you can read your files to overwrite them or use your file as a new template to create new documents"), 
            "BulletsLevel1"
        )
        
        doc.addParagraph("""from unogenerator import ODT
doc=ODT('yourdocument.odt')"""    , "Code")
        
        doc.addParagraph(
            _("ODT without template.") + " " + 
            _("With this mode you can write your files with Libreoffice default styles.") +" " +
            _("If you want to create new ones, you should write them using Libreoffice API code"), 
            "BulletsLevel1"
        )
        
        
        doc.addParagraph("""from unogenerator import ODT
doc=ODT()"""    , "Code")
        
        doc.addParagraph(_("Styles"), "Heading 2")
        doc.addParagraph(
            _("To call default Libreoffice paragraph styles you must use their english name.") + " " + 
            _("You can see their names with this method:"), "Standard"
        )
        doc.addParagraph("""doc.print_styles()"""    , "Code")

        doc.addParagraph(_("Tables"), "Heading 2")
        table_data=[
            [_("Concept"), _("Value"), _("Comment") ], 
            [_("Text"), _("This is a text"), _("Good")], 
            [_("Datetime"), datetime.now(), _("Good")], 
            [_("Date"), date.today(), _("Good")], 
            [_("Float"),  12.121, _("Good")], 
            [_("Currency"), Currency(-12.12, "EUR"), _("Good")], 
            [_("Percentage"), Percentage(1, 3), _("Good")], 
        ]
        
        columnspercentages=[15, 70, 15 ]
        doc.addParagraph(_("We can create tables with diferent font sizes and formats:") + str(columnspercentages), "Standard")
        doc.addTableParagraph(table_data, columnssize_percentages=columnspercentages, style="Table1")
        
        doc.addTableParagraph(table_data, columnssize_percentages=[30, 40,30],  size=6, style="Table1")
        

        doc.addParagraph(_("Lists and numbered lists"), "Heading 2") 
        doc.addParagraph(_("Simple list"), "BulletsLevel1")
        doc.addParagraph(_("Simple list"), "BulletsLevel2")
        doc.addParagraph(_("Simple list"), "BulletsLevel2")
        doc.addParagraph(_("Simple list"), "BulletsLevel1")
        doc.addParagraph(_("Simple list"), "BulletsLevel2")
        doc.addParagraph(_("Simple list"), "BulletsLevel1")


        doc.pageBreak()
        doc.addParagraph(_("Hyperlinks"), "Heading 2")
        
        doc.addString(_("If you want to go to Google, click on this "))
        doc.addStringHyperlink("link",  "https://www.google.com")
        doc.addString(". " + _("That's all folks!"), paragraphBreak=True)
        
        doc.addStringHyperlink("Other link",  "https://www.google.com",  paragraphBreak=True)
        doc.pageBreak()
        
        
        doc.addParagraph(_("HTML code"), "Heading 2")
        doc.addHTMLBlock("<ul><li>This is a html list</li></ul><p style='color:red;'>This is a html paragraph.</p>")
        doc.pageBreak()

        doc.addParagraph(_("Images"), "Heading 2")
        
        l=[]
        l.append( _("This is an 'image as char' example: "))
        l.append(doc.textcontentImage(files('unogenerator') / 'images/crown.png', 1, 1, "AS_CHARACTER", "PRIMERA", linked=False))
        l.append(". "+_("Now I keep writing without problems."))
        doc.addParagraphComplex(l, "Standard")
        
        l=[]
        l.append( _("This is an image loaded from bytes: "))
        
        with open(files('unogenerator') / 'images/crown.png', "rb") as f:
            bytes_crown=f.read()
        
        l.append(doc.textcontentImage(bytes_crown, 1, 1, "AS_CHARACTER", "d", linked=False))
        doc.addParagraphComplex(l, "Standard")

        l=[]
        l.append( _("As you can see, I can reuse it one hundred times. File size will not be increased because I used reference names."))
        for i in range(100):
            l.append(doc.textcontentImage(files('unogenerator') / 'images/crown.png', 0.5,  0.5, "AS_CHARACTER", linked=False))
        doc.addParagraphComplex(l, "Standard")


        doc.addParagraph(_("The next paragraph is generated with the illustration method"), "Standard")
        doc.addImageParagraph([files('unogenerator') / 'images/crown.png']*5, 2.5, 1.5, "Illustration", linked=False)
        
        doc.addParagraph(_("The next paragraph is generated with the illustration method"), "Standard")
        doc.addImageParagraph([files('unogenerator') / 'images/crown.png']*5, 2.5, 1.5, "Illustration", linked=False)
        
        
        doc.addParagraph(_("You can play with image width and height:"), "Standard")
        l=[]
        l.append(doc.textcontentImage(files('unogenerator') / 'images/icons.jpg', None, None,  "AS_CHARACTER", "PRIMERA", linked=False))
        l.append(" Image default size. Height and width are set to None.")
        doc.addParagraphComplex(l, "Standard")
        l=[]
        l.append(doc.textcontentImage(files('unogenerator') / 'images/icons.jpg', 3,  None,  "AS_CHARACTER", "PRIMERA", linked=False))
        l.append(" Image width 3cm of width and height automatically set.")
        doc.addParagraphComplex(l, "Standard")
        l=[]
        l.append(doc.textcontentImage(files('unogenerator') / 'images/icons.jpg',  None, 3,  "AS_CHARACTER", "PRIMERA", linked=False))
        l.append(" Image height 3cm of width and width automatically set.")
        doc.addParagraphComplex(l, "Standard")
        l=[]
        l.append(doc.textcontentImage(files('unogenerator') / 'images/icons.jpg', 3, 3,  "AS_CHARACTER", "PRIMERA", linked=False))
        l.append(" Image width and height set to 3cm.")
        doc.addParagraphComplex(l, "Standard")
        
        
        doc.addParagraph(_("You can trim image border white space when needed:"), "Standard")
        
        l=[]
        l.append("This is a camerawith gray space around it: ")
        l.append(doc.textcontentImage(files('unogenerator') / 'images/Imagewithborder.png', 1, None))
        l.append(". ")
        l.append(_("You can trim that gray space when needed with 'bytes_after_trim_image' method. To use this method you need Imagemagick installed to use 'convert' command. This is the result: "))
        bytes_=commons.bytes_after_trim_image(files('unogenerator') / 'images/Imagewithborder.png', "png")
        l.append(doc.textcontentImage(bytes_, 1, None,))
        doc.addParagraphComplex(l, "Standard")
        
        doc.addParagraph(_("Search and Replace"), "Heading 2")
        doc.addParagraph(_("Below this paragraph is a paragraph with a % REPLACEME % (Without white spaces) text and it's going to be replaced after all document is been generated"), "Standard")
        doc.addParagraph("%REPLACEME%", "Standard")

        doc.pageBreak()
        doc.addParagraph(_("ODS"), "Heading 1")

        doc.addParagraph(_("ODS 'Hello World' example"), "Heading 2")
        doc.addParagraph(_("This is a Hello World example. You'll get the example in ods, xlsx and pdf formats:") ,  "Standard")
        doc.addParagraph("""from unogenerator import ODS_Standard
with ODS_Standard() as doc:
    doc.addCellMergedWithStyle("A1:E1", "Hello world", style="BoldCenter")
    doc.save("hello_world.ods")
    doc.export_xlsx("hello_world.xlsx")
    doc.export_pdf("hello_world.pdf")"""    , "Code")
        
        doc.find_and_replace("%REPLACEME%", _("This paragraph was set at the end of the code after a find and replace command."))
        doc.paragraphBreak()
        doc.addParagraph(_("This paragraph was set after replacement."), "Standard")
        doc.pageBreak()
        doc.addParagraph(_("This paragraph was set after a page break."), "Standard")
        doc.pageBreak("Landscape")
        doc.addParagraph(_("This paragraph was set after a page break with Landscape style."), "Standard")
        
        
    #    doc.find_and_delete_until_the_end_of_document("This paragraph was set after replacement.")

        doc.addParagraph(_("This is a pair of brackets ()."), "Standard")
        doc.findall_and_replace(" ().", " )(.", True)
        
        doc.addParagraph(_("NOW)("), "Standard")
        doc.addParagraph(_("This is a set of symbols: .,:;?ºª-/()."), "Standard")
        doc.findall_and_replace(".,:;?ºª-/().", ".,:;?ºª-/(). REPLACED", True)
        doc.addParagraph(_("NOW)("), "Standard")
        
        
        
        doc.save(f"unogenerator_documentation_{language}.odt")
        doc.export_docx(f"unogenerator_documentation_{language}.docx")
        doc.export_pdf(f"unogenerator_documentation_{language}.pdf")

    r= _("unogenerator_documentation_{0}.ods took {1} in {2}").format(language, datetime.now()-doc.start, doc.server.port)
    info(r)
    return r




from argparse import ArgumentParser, RawTextHelpFormatter
from colorama import init as colorama_init
from copy import deepcopy
from datetime import datetime, timedelta
from gettext import translation
from pydicts import lod
from importlib.resources import files
from humanize import naturalsize
from psutil import process_iter
from unogenerator import __version__
from unogenerator.commons import argparse_epilog, addDebugSystem, green, red, yellow, white
from os import system, scandir
from pydicts.percentage import Percentage
from time import sleep

try:
    t=translation('unogenerator', files("unogenerator") / 'locale')
    _=t.gettext
except:
    _=str


    
    
def monitor():    
    colorama_init()
    parser=ArgumentParser(
        description=_('UnoGenerator monitor'), 
        epilog=argparse_epilog(), 
        formatter_class=RawTextHelpFormatter
    )
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--debug', help=_("Debug program information"), choices=["DEBUG","INFO","WARNING","ERROR","CRITICAL"], default="ERROR")
    parser.add_argument('--refresh', help=_('Seconds until next refresh'), action="store",  type=int,  default=1)
    parser.add_argument('--seconds', help=_('Seconds to color Last CPU variation'), action="store",  type=int,  default=60)
    
    args=parser.parse_args()

    addDebugSystem(args.debug)
    command_monitor(args.seconds, args.refresh)

## @param restart boolean. To restart unogenerator server when idle and used memory above recommended memory
## @param recommended. Integer. Recomended memory in Mb
def command_monitor(seconds, refresh):
    last_dod_=None
    while True:
        
        
        
        system("reset")

        print (_("Unogenerator version: {0}. Hora actual: {1}. Refresh time: {2} seconds").format(yellow(__version__), white(datetime.now()),  green(refresh)))
        print ()

      
        
        
        
        directorios_deben=set()
        dod_={}
        for p in process_iter(['name','cmdline', 'pid']): 
            d={}
            try:
                if p.info['name']=='soffice.bin':
                    if  p.info['cmdline'] is not None and 'file:///tmp/unogenerator'  in ' '.join(p.info['cmdline']):
                        d["pid"]=p.pid
                        d["port"]=int(p.info['cmdline'][1].split("unogenerator")[1])
                        directorios_deben.add(f"unogenerator{d['port']}")
                        d["mem"]=naturalsize(p.memory_info().rss)
                        d["status"]=p.status()
                        d["duration"]=datetime.now()-datetime.fromtimestamp(p.create_time())
                        d["cpu_percentage"]=Percentage(p.cpu_percent(interval=0.01), 100)
                        connections=len(p.connections())
                        d["conexiones"]=connections
                        
                        if last_dod_ is None: #Primera ronda
                            d["last_cpu_percentage_datetime"]=datetime.now()
                            d["Last CPU"]=timedelta(seconds=0)
                        else: #Siguientes rondas
                            #Actualiza cpu_percentage
                            if d["cpu_percentage"].value>0:
                                d["last_cpu_percentage_datetime"]=datetime.now()
                                d["Last CPU"]=timedelta(seconds=0)
                            else: #No se ha utilizado cpu
                                if p.pid in last_dod_ and "last_cpu_percentage_datetime" in last_dod_[p.pid]:
                                    d["last_cpu_percentage_datetime"]=last_dod_[p.pid]["last_cpu_percentage_datetime"]
                                    d["Last CPU"]=datetime.now()-last_dod_[p.pid]["last_cpu_percentage_datetime"]
                                else:
                                    d["last_cpu_percentage_datetime"]=datetime.now()
                                    d["Last CPU"]=timedelta(seconds=0)

                        dod_[p.pid]=d
            except:
                pass
        last_dod_=deepcopy(dod_)
        
        result_dod=deepcopy(dod_)
        result_lod=lod.lod_order_by(result_dod.values(), "duration",  reverse=True)
        lod.lod_remove_key(result_lod,"last_cpu_percentage_datetime")
        
        #Change color on Last CPU
        for d in result_lod:
            d["duration"]=red(d["duration"] )if d["duration"].total_seconds()>seconds else green(d["duration"])
            d["Last CPU"]=red(d["Last CPU"] )if d["Last CPU"].total_seconds()>seconds else green(d["Last CPU"])
        
        lod.lod_print(result_lod)
        
        print(green(_("Active soffice.bin processes number:")),  white(len(result_lod)))
        
        directorios_existen=set()
        for entry in scandir("/tmp"):
            if entry.is_dir() and entry.name.startswith("unogenerator"):
                directorios_existen.add(entry.name)
        
        print(green(_("Temporal directories with asociated soffice.bin process: ")),  yellow(list(directorios_existen-directorios_deben)))
       
        sleep(float(refresh))    
    
def cleaner():    
    colorama_init()
    parser=ArgumentParser(
        description=_('UnoGenerator cleaner'), 
        epilog=argparse_epilog(), 
        formatter_class=RawTextHelpFormatter
    )
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--debug', help=_("Debug program information"), choices=["DEBUG","INFO","WARNING","ERROR","CRITICAL"], default="ERROR")
    args=parser.parse_args()

    addDebugSystem(args.debug)
    command_cleaner()
    
## @param restart boolean. To restart unogenerator server when idle and used memory above recommended memory
## @param recommended. Integer. Recomended memory in Mb
def command_cleaner():
    system("killall -9 soffice.bin")
    system("rm -Rf /tmp/unogenerator*")



from argparse import ArgumentParser,  RawTextHelpFormatter
from gettext import translation
from os import path,  makedirs
from polib import POEntry,  POFile, pofile
from subprocess import run
from importlib.resources import files
from unogenerator.commons import argparse_epilog, red
from unogenerator import ODT, __version__


try:
    t=translation('unogenerator', files("unogenerator") / 'locale')
    _=t.gettext
except:
    _=str

def run_check(command, shell=False):
    p=run(command, shell=shell, capture_output=True);
    if p.returncode!=0:
        print(f"Error en comando. {command}")
        print("STDOUT:")
        print(p.stdout.decode('utf-8'))
        print("STDERR:")
        print(p.stderr.decode('utf-8'))
        print("Saliendo de la instalación")
        exit(2)

            
def translation():
    parser=ArgumentParser(prog='unogenerator', description=_('Generate a po and pot file from ODF'), epilog=argparse_epilog(), formatter_class=RawTextHelpFormatter)
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('--from_language', action='store', help=_('Language to translate from. Example codes: es, fr, en, md'), required=True, metavar="CODE")
    parser.add_argument('--to_language', action='store', help=_('Language to translate to. Example codes: es, fr, en, md'), required=True,  metavar="CODE")
    parser.add_argument('--input', action='append', help=_('Files to translate. You can set several files.'), required=True,  metavar="FILE")
    parser.add_argument('--output_directory', action='store', help=_('Output directory with results and catalogues. Default: ./unogenerator_translation/') , metavar="FILE",  default="./unogenerator_translation/")
    parser.add_argument('--undetected', action='append', help=_('Undetected strings to append to translation. Default: []'), default=[])
    parser.add_argument('--fake', action='store_true', help=_('Sets a fake translation to all strings. Default: False'), default=False)
    parser.add_argument('--pdf', action='store_true', help=_('Creates translated file in pdf. Default: False'), default=False)
    args=parser.parse_args()
    
    command_translation(args.from_language, args.to_language, args.input, args.output_directory, args.undetected, args.fake, args.pdf)

def same_entries_to_ocurrences(l):
    l= sorted(l, key=lambda x: (x[0], x[1], x[2], x[3]))
    r=[]
    for filename, type, number,  position,  text in l:
        r.append((filename, f"{type}#{number}#{position}"))
    return r

    
def getEntriesFromDocument(filename):
        r=[]
        doc=ODT(filename)
        #Extract strings from metadata
        metadata=doc.getMetadata()
        for key,  value in metadata.items():
            if value.__class__.__name__=="str" and len(value)>0:
                r.append((filename, f"Metadata_{key}",  0,  0, value))

        #Extract strings from paragraphs
        r=r+entries_from_paragraph_enumeration("Paragraph", doc.cursor.Text.createEnumeration(), filename)

        for style in doc.getPageStylesObjects():            
            #Extract strings from headers
            ht=style.HeaderText
            if ht is None:
                continue
            r=r+entries_from_paragraph_enumeration("HeaderParagraph", ht.createEnumeration(), filename)
            #Extract strings from foot
            ft=style.FooterText
            if ft is None:
                continue
            r=r+entries_from_paragraph_enumeration("FooterParagraph", ft.createEnumeration(), filename)
        doc.close()
        return r
        
def entries_from_paragraph_enumeration(title, enumeration, filename):
        r=[]
        for i,  par in enumerate(enumeration):
#            print(i, par, dir(par))
            if  par.supportsService("com.sun.star.text.Paragraph") :
                for position, element in enumerate(par.createEnumeration()):
                    text_=element.getString()
                    if text_ !="" and text_!=" " and text_!="  ":
                        entry=(filename, title,  i,  position, text_)
#                        print(entry)
                        r.append(entry)
            elif  par.supportsService("com.sun.star.text.TextTable") :
#                print("AQUQI", par, dir(par), par.getColumns().Count, par.getRows().Count)
                for column in range(par.getColumns().Count):
                    for row in range(par.getRows().Count):
                        try:
                            cell=par.getCellByPosition(column, row)
                            r=r+entries_from_paragraph_enumeration(f"{title}Table,paragraph{i},column{column},row{row}", cell.createEnumeration(), filename)
                        except:
                            print (_("Perhaps a merged string"))
                            continue
        return r
            
        

            

def generate_pot_file(potfilename, set_strings, entries):
    file_pot = POFile()
    file_pot.metadata = {
        'Project-Id-Version': '1.0',
        'Report-Msgid-Bugs-To': 'you@example.com',
        'POT-Creation-Date': '2007-10-18 14:00+0100',
        'PO-Revision-Date': '2007-10-18 14:00+0100',
        'Last-Translator': 'you <you@example.com>',
        'Language-Team': 'English <yourteam@example.com>',
        'MIME-Version': '1.0',
        'Content-Type': 'text/plain; charset=utf-8',
        'Content-Transfer-Encoding': '8bit',
    }
    for s in set_strings:
        same_entries=[] #Join seame text entries
        for filename, type, number, position, string_ in entries:
            if string_==s:
                same_entries.append((filename, type, number, position, string_))

        entry = POEntry(
            msgid=s,
            msgstr='', 
            occurrences=same_entries_to_ocurrences(same_entries)
        )
        file_pot.append(entry)
    file_pot.save(potfilename)

def command_translation(from_language, to_language, input, output_directory="./unogenerator_translation/", undetected_strings=[], fake=False, pdf=False):   
    makedirs(output_directory, exist_ok=True)
    makedirs(f"{output_directory}/{to_language}", exist_ok=True)
    
    pot=f"{output_directory}/catalogue.pot"
    po=f"{output_directory}/{to_language}/{to_language}.po"
        
    entries=[]#List of (filename,"type", numero, posicion) type=Paragraph, numero=numero parrafo y posición orden dentro del parrafo
    print(_("Extracting strings from:"))
    
    goodinput=[]#To remove bad files and dont make errors deleting in list
    for filename in input:
        if filename.upper()[-3:]!="ODT":
            print(f"   - {filename}. " + red(_("Only ODT files are supported")))
            continue
        print(f"   - {filename}")
        goodinput.append(filename)
        entries=entries+getEntriesFromDocument(filename)
    input=goodinput
        
        
    # Set with distinct strings of all entries
    set_strings=set()
    for filename, type, number, position, string_ in entries:
        set_strings.add(string_)

    #Generate pot file
    print(_("Generating {0}").format(pot))
    generate_pot_file(pot, set_strings, entries)
    
    #Merging pot file to po (new or not)
    if path.exists(po) is False:
        run_check(["msginit", "-i", pot,  "-o", po])
    run_check(["msgmerge","-N", "--no-wrap","-U", "-F", po, pot])
    print(_("   - {0} different strings detected").format(len(set_strings)))
    
    # Creates a dictionary of translations
    dict_po={}
    for i, entry in enumerate(pofile(po)):
        if fake is True:
            dict_po[entry.msgid]=f"{{{entry.msgid}}}"
        else:
            if entry.msgstr == "":
                dict_po[entry.msgid]=entry.msgid
            else:
                dict_po[entry.msgid]=entry.msgstr
    
    print(_("Translating files to:"))
    for filename_input in input:
        output=f"{output_directory}/{to_language}/{path.basename(filename_input)}"
        print(f"   + {output}")
        write_translation(filename_input, output,  dict_po, entries, pdf)
            
            
def write_translation(original, filename, dict_po, entries, pdf):
    doc=ODT(original)
    search_descriptor=None
    replaced=0
    #Replaces all strings generic
    for filename_, type, number,  position,  text in entries:
        if filename_==path.basename(filename):
            #Translate metadata
            if type=="Metadata_Author":
                doc.setMetadata(author=dict_po[text])
                replaced=replaced+1
            elif type=="Metadata_Description":
                doc.setMetadata(description=dict_po[text])
                replaced=replaced+1
            elif type=="Metadata_Title":
                doc.setMetadata(title=dict_po[text])
                replaced=replaced+1
            elif type=="Metadata_Subject":
                doc.setMetadata(subject=dict_po[text])
                replaced=replaced+1
            else:#Translate rest of entries
                search_descriptor=doc.find_and_replace(text, dict_po[text], search_descriptor)
                if search_descriptor is not None:
                    replaced=replaced+1
                
    
    print(f"      - Translated {replaced} strings")
    doc.save(filename)
    if pdf is True:
        print("      - PDF translation generated")
        doc.export_pdf(filename[:-4] +".pdf")
    doc.close()
        




from colorama import Style, Fore

def can_import_uno(output=True):
    """
        Check if uno module (from LibreOffice package) can be imported
        
        This should be used by external programs to avoid crashes when uno.py modules is not found
        
        Parameters:
            - output -> bool: Show fail text
    """
    try:
        from uno import Any
        Any    
    except ImportError:
        if output is True:
            e="""This exception is due to you haven't been able to import uno.py module.
        
This module belongs to libreoffice package and you won't find it in systems like pip, poetry ....

To fix it, you must try:
    - Install libreoffice and its python modules
    - Check your python version is the same as the libreoffice compiled module of python
    - If you're using a  python virtual environment you must edit pyvenv.cfg and set:   include-system-site-packages = true
"""
            print(Style.BRIGHT + Fore.RED + str(e) + Style.RESET_ALL)
        return False
    return True



from unogenerator import __version__
from os import system
from sys import argv


def test():
    system("pytest -W error")
    
def coverage():
    system("coverage run --omit='*uno.py' -m pytest && coverage report && coverage html")

def translate():
        system("xgettext -L Python --no-wrap --no-location --from-code='UTF-8' -o unogenerator/locale/unogenerator.pot  unogenerator/*.py")
        system("msgmerge -N --no-wrap -U unogenerator/locale/es.po unogenerator/locale/unogenerator.pot")
        system("msgfmt -cv -o unogenerator/locale/es/LC_MESSAGES/unogenerator.mo unogenerator/locale/es.po")
        system("msgfmt -cv -o unogenerator/locale/en/LC_MESSAGES/unogenerator.mo unogenerator/locale/en.po")
        system("msgfmt -cv -o unogenerator/locale/fr/LC_MESSAGES/unogenerator.mo unogenerator/locale/fr.po")
        system("msgfmt -cv -o unogenerator/locale/ro/LC_MESSAGES/unogenerator.mo unogenerator/locale/ro.po")

    
def documentation():
        system("unogenerator_demo --create")
        system("cp -f unogenerator_documentation_en.odt doc/")
        system("cp -f unogenerator_documentation_en.pdf doc/")
        system("cp -f unogenerator_documentation_es.odt doc/")
        system("cp -f unogenerator_documentation_es.pdf doc/")
        system("cp -f unogenerator_example_en.ods doc/")
        system("cp -f unogenerator_example_en.pdf doc/")
        system("cp -f unogenerator_example_es.ods doc/")
        system("cp -f unogenerator_example_es.pdf doc/")
        system("unogenerator_demo --remove")

def release():
    print("""Nueva versión:
  * Cambiar la versión y la fecha en __init__.py
  * Cambiar la versión en pyproject.toml
  * Ejecutar otra vez poe release
  * git checkout -b unogenerator-{0}
  * pytest
  * unogenerator_demo --benchmark
  * Modificar el Changelog en README.md
  * poe coverage con pyvenv systempackages=false
  * poe coverage con pyvenv systempackages=true
  * poe translate
  * linguist
  * poe translate
  * poe documentation
  * git commit -a -m 'unogenerator-{0}'
  * git push
  * Hacer un pull request con los cambios a main
  * Hacer un nuevo tag en GitHub
  * git checkout main
  * git pull
  * poetry build
  * poetry publish --username --password  
  * Crea un nuevo ebuild de UNOGENERATOR Gentoo con la nueva versión
  * Subelo al repositorio del portage

""".format(__version__))

def docker_build():
    system("docker build --tag turulomio/unogenerator:latest .")

def docker():
    system(f"docker  run  -p 127.0.0.1:2002:2002 -it {argv[1]}")



class CoordException(Exception):
    """
        Raised when Coord is wrong
    """
    pass
    
class RangeException(Exception):
    """
        Raised when Coord is wrong
    """
    pass

class UnogeneratorException(Exception): 
    pass
  
